# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""A library to assess adjustable demographically fair Machine Learning (ML) systems."""

from .plot import pareto_plot, radar_chart
from .scores import Scores
from .solutions import Solutions
from .utils import make_table

__all__ = [
    "Scores",
    "Solutions",
    "make_table",
    "pareto_plot",
    "radar_chart",
]



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
import pathlib

from fairical.scores import Scores
from fairical.solutions import Solutions


def test_a_priori_metadata(datadir: pathlib.Path):
    scores = Scores.load(str(datadir / "data" / "sample" / "system_1.json"))
    prior_solutions = Solutions.load(str(datadir / "data" / "uc-1" / "system_2.json"))
    prior_solutions_nds = prior_solutions.non_dominated_solutions()[0]

    metric = list(prior_solutions.points.keys())

    solutions = scores.solutions_a_priori(metric, prior_solutions, dominated=False)

    expected_thr_ids_combinations = list(
        zip(
            prior_solutions_nds.metadata["thresholds"],
            prior_solutions_nds.metadata["identifiers"],
        )
    )
    thr_ids_combinations = list(
        zip(solutions.metadata["thresholds"], solutions.metadata["identifiers"])
    )

    for combination in thr_ids_combinations:
        assert combination in expected_thr_ids_combinations


def test_tabulate():
    sol = Solutions.fromarray(
        [[0.1, 0.9], [0.2, 0.2], [0.3, 0.3]], ("eod+race", "acc"), [0.6, 0.2, 0.3]
    )

    table = sol.tabulate()

    assert "(eod+race, acc)" in table
    assert "thresholds" in table
    assert "identifiers" in table
    assert "(0.1, 0.9)" in table
    assert "(0.2, 0.2)" in table
    assert "(0.3, 0.3)" in table
    assert "0.6" in table
    assert "0.2" in table
    assert "0.3" in table
    assert table.count("model-0") == 3



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tools for command-line applications."""

import csv
import logging
import pathlib
import shutil
import typing

import click
import compact_json
import numpy

from ..scores import Scores
from ..solutions import Solutions

logger = logging.getLogger(__name__)


def prepare_and_backup(path: pathlib.Path) -> None:
    """Ensure parent directory exists and back-up copies.

    This function will check that the directory leading to a file path exists and will
    created it otherwise.  It will also check if the file does not already exists, and
    back it up otherwise.

    Parameters
    ----------
    path
        The full path of the file to ensure the parent directory exists and that it is
        properly backed-up if necessary.
    """

    path.parent.mkdir(parents=True, exist_ok=True)
    if path.exists():
        backup = path.parent / (path.name + "~")
        shutil.copy(path, backup)


def save_json_with_backup(path: pathlib.Path, data: dict) -> None:
    """Save a dictionary into a JSON file with path checking and backup.

    This function will save a dictionary into a JSON file.  It will check
    the existence of the directory leading to the file and create it if
    necessary.  If the file already exists on the destination folder, it is
    backed-up before a new file is created with the new contents.

    Parameters
    ----------
    path
        The full path where to save the JSON data.
    data
        The data to save on the JSON file.
    """

    formatter = compact_json.Formatter()
    # only only 2 indent spaces for further levels
    formatter.indent_spaces = 2
    # controls how much nesting can happen
    formatter.max_inline_complexity = 2
    # controls the maximum line width (has priority over nesting)
    formatter.max_inline_length = 88
    # remove any trailing whitespaces
    formatter.omit_trailing_whitespace = True

    prepare_and_backup(path)

    formatter.dump(data, str(path))


def save_csv_with_backups(path: pathlib.Path, data: dict[str, typing.Any]):
    """Save a dictionary into a CSV file with path checking and backup.

    This function will save a dictionary into a CSV file.  It will check
    the existence of the directory leading to the file and create it if
    necessary.  If the file already exists on the destination folder, it is
    backed-up before a new file is created with the new contents.

    Parameters
    ----------
    path
        The full path where to save the CSV data.
    data
        The data to save on the CSV file.
    """

    prepare_and_backup(path)

    with path.open("w") as csv_file:
        csv_writer = csv.writer(csv_file)
        csv_writer.writerow(data.keys())
        for d in data.values():
            csv_writer.writerow(d)


def validate_scores_json(
    ctx: click.Context, param: click.Parameter, value: typing.Sequence[pathlib.Path]
) -> list[
    tuple[pathlib.Path, Scores]
    | tuple[tuple[pathlib.Path, Scores], tuple[pathlib.Path, Solutions]]
]:
    """
    Validate one or more JSON files against a predefined data model.

    This function is intended to be used as a Click argument callback. It opens the
    given file path, parses its JSON content, and validates it against the library data
    model. If any validation error occurs, a `click.BadParameter` is raised to inform
    the user.

    Parameters
    ----------
    ctx
        The Click execution context.
    param
        The parameter that triggered the callback.
    value
        Path to the JSON files to load and validate.

    Returns
    -------
        The parsed and validated JSON content as a list, where each element is either a
        a tuple mapping filename to scores, or a tuple containing two entries, the first
        mapping a filename to scores and a second (to be used for selecting *a priori*
        thresholds and systems) mapping a filename to pre-calculated solutions.

    Raises
    ------
    click.BadParameter
        If the file cannot be read, contains invalid JSON, or fails data model
        validation.
    """

    retval: list[
        tuple[pathlib.Path, Scores]
        | tuple[tuple[pathlib.Path, Scores], tuple[pathlib.Path, Solutions]]
    ] = []

    def _load_scores(v):
        try:
            assert v.stem not in [k[0] for k in retval]
            logger.info(f"Loading scores for system `{v}`...")
            return Scores.load(v)

        except AssertionError as e:
            raise click.BadParameter(f"The same path was passed more than once: {e}")
        except __import__("pydantic").ValidationError as e:
            raise click.BadParameter(f"Score data model validation failed: {e}")

    def _load_solutions(v):
        try:
            logger.info(f"Loading solutions for system `{v}`...")
            return Solutions.load(v)
        except __import__("pydantic").ValidationError as e:
            raise click.BadParameter(f"Solution data model validation failed: {e}")

    if isinstance(value, pathlib.PosixPath):
        value = (value,)

    for v in value:
        system_groups = [pathlib.Path(p) for p in str(v).split("@")]
        scores = _load_scores(system_groups[0])

        if len(system_groups) == 1:
            retval.append((system_groups[0], scores))

        elif len(system_groups) == 2:
            retval.append(
                (
                    (system_groups[0], scores),
                    (system_groups[1], _load_solutions(system_groups[1])),
                )
            )

        else:
            raise click.BadParameter(f"Too many systems specified in group: {v}")

    return retval


def validate_solutions_json(
    ctx: click.Context, param: click.Parameter, value: typing.Sequence[pathlib.Path]
) -> dict[str, Solutions]:
    """
    Validate one or more JSON files against a predefined data model.

    This function is intended to be used as a Click argument callback. It opens the
    given file path, parses its JSON content, and validates it against the library data
    model. If any validation error occurs, a `click.BadParameter` is raised to inform
    the user.

    Parameters
    ----------
    ctx
        The Click execution context.
    param
        The parameter that triggered the callback.
    value
        Path to the JSON files to load and validate.

    Returns
    -------
        The parsed and validated JSON content as a dictionary, where keys represent the
        basename of files without extension.

    Raises
    ------
    click.BadParameter
        If the file cannot be read, contains invalid JSON, or fails data model
        validation.
    """

    retval: dict[str, Solutions] = {}

    if isinstance(value, pathlib.PosixPath):
        value = (value,)

    for v in value:
        try:
            assert v.stem not in retval
            logger.info(f"Loading solutions for system `{v}`...")
            retval[v.stem] = Solutions.load(v)
        except AssertionError as e:
            raise click.BadParameter(f"The same path was passed more than once: {e}")
        except __import__("pydantic").ValidationError as e:
            raise click.BadParameter(f"Solution data model validation failed: {e}")

    return retval


def validate_metrics(
    ctx: click.Context, param: click.Parameter, value: typing.Sequence[str]
) -> tuple[str, ...]:
    """Validate a user defined metric for support.

    Parameters
    ----------
    ctx
        The Click execution context.
    param
        The parameter that triggered the callback.
    value
        The value to validate.

    Returns
    -------
        The validated metrics.

    Raises
    ------
    click.BadParameter
        If one or more of the provided metrics do not validate correctly.
    """
    from .. import metrics

    invalid = []
    for name in value:
        try:
            metrics.parse_metric(name)
        except ValueError:
            invalid.append(f"`{name}`")

    if invalid:
        raise click.BadParameter(f"invalid metric names: {', '.join(invalid)}")

    return tuple(value)


def validate_thresholds(
    ctx: click.Context, param: click.Parameter, value: int | None
) -> None | list[float]:
    """Validate a threshold setup.

    Parameters
    ----------
    ctx
        The Click execution context.
    param
        The parameter that triggered the callback.
    value
        The value to validate.

    Returns
    -------
        The validated threshold, as our API likes it.

    Raises
    ------
    click.BadParameter
        If the threshold cannot be validated.
    """

    if value is None:
        return value

    return typing.cast(list[float], list(numpy.linspace(0, 1, value, dtype=float)))


def verbosity_option(
    logger: logging.Logger,
    short_name: str = "v",
    name: str = "verbose",
    dflt: int = 0,
    **kwargs: typing.Any,
) -> typing.Callable[..., typing.Any]:
    """Click-option decorator that adds a ``-v``/``--verbose`` option to a cli.

    This decorator adds a click option to your CLI to set the log-level on a
    provided :py:class:`logging.Logger`.  You must specifically determine the
    logger that will be affected by this CLI option, via the ``logger`` option.

    .. code-block:: python

       @verbosity_option(logger=logger)

    The verbosity option has the "count" type, and has a default value of 0.
    At each time you provide ``-v`` options on the command-line, this value is
    increased by one.  For example, a CLI setting of ``-vvv`` will set the
    value of this option to 3.  This is the mapping between the value of this
    option (count of ``-v`` CLI options passed) and the log-level set at the
    provided logger:

    * 0 (no ``-v`` option provided): ``logger.setLevel(logging.ERROR)``
    * 1 (``-v``): ``logger.setLevel(logging.WARNING)``
    * 2 (``-vv``): ``logger.setLevel(logging.INFO)``
    * 3 (``-vvv`` or more): ``logger.setLevel(logging.DEBUG)``


    Arguments:

        logger: The :py:class:`logging.Logger` to be set.

        short_name: Short name of the option.  If not set, then use ``v``

        name: Long name of the option.  If not set, then use ``verbose`` --
            this will also become the name of the contextual parameter for click.

        dlft: The default verbosity level to use (defaults to 0).

        **kwargs: Further keyword-arguments to be forwarded to the underlying
            :py:func:`click.option`


    Returns
    -------
        A callable, that follows the :py:mod:`click`-framework policy for
        option decorators.  Use it accordingly.
    """

    def custom_verbosity_option(f):
        def callback(ctx, _, value):
            ctx.meta[name] = value
            log_level: int = {  # type: ignore
                0: logging.ERROR,
                1: logging.WARNING,
                2: logging.INFO,
                3: logging.DEBUG,
            }[value]

            # one‐time handler setup
            if not getattr(logger, "_verbosity_configured", False):
                handler = logging.StreamHandler()
                handler.setFormatter(
                    logging.Formatter(
                        "%(levelname)s %(name)s: %(message)s",
                    )
                )
                logger.addHandler(handler)
                logger._verbosity_configured = True  # type: ignore[attr-defined]

            logger.setLevel(log_level)
            logger.debug(f'Level of Logger("{logger.name}") was set to {log_level}')

            return value

        return click.option(
            f"-{short_name}",
            f"--{name}",
            count=True,
            type=click.IntRange(min=0, max=3, clamp=True),
            default=dflt,
            show_default=True,
            help=(
                f"Increase the verbosity level from 0 (only error and "
                f"critical) messages will be displayed, to 1 (like 0, but adds "
                f"warnings), 2 (like 1, but adds info messags), and 3 (like 2, "
                f"but also adds debugging messages) by adding the --{name} "
                f"option as often as desired (e.g. '-vvv' for debug)."
            ),
            callback=callback,
            **kwargs,
        )(f)

    return custom_verbosity_option



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy

from fairical.solutions import Solutions


def test_single_pareto_point() -> None:
    """Only one solution should survive when it dominates all others."""

    sol = Solutions.fromarray(
        [[0.1, 0.9], [0.2, 0.2], [0.3, 0.3]], ("eod+race", "acc"), [0.6, 0.2, 0.3]
    )
    nds, ds = sol.non_dominated_solutions()

    front = numpy.asarray(nds)
    back = numpy.asarray(ds)

    assert front.shape == (1, sol.n_metrics())
    numpy.testing.assert_allclose(front[0], [0.1, 0.9])

    assert back.shape == (2, sol.n_metrics())
    numpy.testing.assert_allclose(back, numpy.array([[0.2, 0.2], [0.3, 0.3]]))

    # check ONVG and ONVGR
    indicators = sol.indicators()
    assert indicators["onvg"] == 1.0
    assert indicators["onvgr"] == 1 / 3


def test_two_non_dominated_points() -> None:
    """Two solutions mutually non‑dominated should both be returned."""

    sol = Solutions.fromarray(
        [[0.1, 0.8], [0.2, 0.9], [0.3, 0.3]], ("eod+age", "f1"), [0.6, 0.4, 0.3]
    )
    nds, ds = sol.non_dominated_solutions()

    front = numpy.asarray(nds)
    back = numpy.asarray(ds)

    assert front.shape == (2, sol.n_metrics())
    numpy.testing.assert_allclose(front, numpy.array([[0.1, 0.8], [0.2, 0.9]]))

    assert back.shape == (1, sol.n_metrics())
    numpy.testing.assert_allclose(back[0], [0.3, 0.3])

    # check ONVG and ONVGR
    indicators = sol.indicators()
    assert indicators["onvg"] == 2.0
    assert indicators["onvgr"] == 2 / 3


def test_duplicate_points() -> None:
    """Identical optimal points are all kept; duplicates aren't filtered."""

    sol = Solutions.fromarray(
        [[0.1, 0.9], [0.1, 0.9], [0.2, 0.2]], ("eod+gender", "acc"), [0.5, 0.5, 0.2]
    )
    nds, ds = sol.non_dominated_solutions()

    front = numpy.asarray(nds)
    back = numpy.asarray(ds)

    assert front.shape == (2, sol.n_metrics())
    numpy.testing.assert_allclose(front, numpy.array([[0.1, 0.9], [0.1, 0.9]]))

    assert back.shape == (1, sol.n_metrics())
    numpy.testing.assert_allclose(back[0], [0.2, 0.2])

    # check ONVG and ONVGR
    indicators = sol.indicators()
    assert indicators["onvg"] == 2.0
    assert indicators["onvgr"] == 2 / 3


def test_deduplicate_points() -> None:
    """Identical optimal points are all kept; duplicates are filtered."""

    sol = Solutions.fromarray(
        [[0.1, 0.9], [0.1, 0.9], [0.2, 0.2]], ("eod+race", "bal_acc"), [0.5, 0.5, 0.2]
    )
    nds, ds = sol.deduplicate().non_dominated_solutions()

    front = numpy.asarray(nds)
    back = numpy.asarray(ds)

    assert front.shape == (1, sol.n_metrics())
    numpy.testing.assert_allclose(front[0], [0.1, 0.9])

    assert back.shape == (1, sol.n_metrics())
    numpy.testing.assert_allclose(back[0], [0.2, 0.2])

    # check ONVG and ONVGR
    indicators = sol.deduplicate().indicators()
    assert indicators["onvg"] == 1.0
    assert indicators["onvgr"] == 0.5


def test_three_objective_case() -> None:
    """Verify correct Pareto extraction with three objectives."""

    sol = Solutions.fromarray(
        [
            [0.1, 0.1, 0.5],  # non‑dominated
            [0.5, 0.1, 0.9],  # non‑dominated
            [0.1, 0.5, 0.9],  # non‑dominated
            [0.3, 0.3, 0.4],  # dominated by (0.1,0.1,0.5)
            [0.2, 0.2, 0.8],  # non-dominated
        ],
        ("eod+race", "eod+gender", "roc_auc"),
        [0.3, 0.4, 0.6, 0.35, 0.7],
    )
    nds, ds = sol.non_dominated_solutions()

    front = numpy.asarray(nds)
    back = numpy.asarray(ds)

    assert front.shape == (4, sol.n_metrics())
    # Ensure the dominated point is NOT present
    assert not any(numpy.allclose(p, [0.3, 0.3, 0.4]) for p in front)

    assert back.shape == (1, sol.n_metrics())
    # Ensure the dominated point is there
    numpy.testing.assert_allclose(back[0], [0.3, 0.3, 0.4])

    # check ONVG and ONVGR
    indicators = sol.indicators()
    assert indicators["onvg"] == 4.0
    assert indicators["onvgr"] == 0.8



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Define the basic solution data model and functionality."""

import logging
import pathlib
import typing

logger = logging.getLogger(__name__)

import numpy
import numpy.typing
import pydantic
import pymoo.config
import pymoo.indicators.hv
import pymoo.util.nds.non_dominated_sorting
import sklearn.cluster
import tabulate

from . import metrics, utils

# Supress pymoo warnings
pymoo.config.Config.show_compile_hint = False


def _hypervolume(
    nds: numpy.typing.NDArray[numpy.float64],
    nadir_point: numpy.typing.ArrayLike,
) -> float:
    r"""Calculate hyper-volume(hv) of non-dominated solutions.

    This method calculates convergence-diversity based on union of hypervolumes created
    by points from candidate solution set (non-dominated) and Nadir point. Higher is
    better.

    This indicator evaluates how the solution set covers the metric space in terms of
    diversity and proximity to the ideal. HV is formulated as:

    .. math::
        HV(S) = VOL\left(\bigcup_{\substack{x \in S \\ x \prec r}} \prod_{i=1}^{N}[x^{i},r^{i}]\right)

    Where :math:`x` is the solution set and :math:`r` is the Nadir point

    Parameters
    ----------
    nds
        Array with shape ``(n_solutions, n_objectives)`` containing the non-dominated
        solutions (NDS).  **All objectives are assumed to be minimised.**  If your
        problem involves maximisation, multiply the corresponding columns by ``-1``
        before calling.
    nadir_point
        Nadir point with shape ``(n_objectives,)``. Must **not** dominate any point in
        ``nds``.

    Returns
    -------
        The hyper-volume value.

    Raises
    ------
    ValueError
        If ``nadir_point`` is not an 1-D array or does not match with ``nds``.
    """

    nadir_point_ = numpy.asarray(nadir_point, dtype=float)

    # validation
    if nadir_point_.ndim != 1:
        raise ValueError("nadir_point must be a 1‑D array of shape (n_objectives,)")

    if nds.ndim != 2:
        raise ValueError("nds.ndim should be 2")

    if nds.shape[1] != nadir_point_.shape[0]:
        raise ValueError(
            f"nds.shape[1] ({nds.shape[1]}) should match nadir_point.shape[0]"
        )

    ind = pymoo.indicators.hv.Hypervolume(ref_point=nadir_point_)
    return ind._do(nds)


def _uniform_distribution(
    nds: numpy.typing.NDArray[numpy.float64], shared_dist: float = 0.01
) -> float:
    r"""Calculate Uniform Distribution (UD).

    This method calculates diversity/distribution based on distribution of
    non-dominated solutions found on trade-off surface. Lower is better.

    This indicator evaluates how uniform the solution set is spanned in the
    metric space based on an upper-bound distance, :math:`\sigma`. UD is formulated as:

    .. math::

       UD(S,\sigma)=\frac{1}{1+D_{nc}(S, \sigma)}

    Where

    .. math::

        D_{nc}(S,\sigma)=\sqrt{\frac{1}{|X_n|-1} \sum_{i=1}^{|X_n|} \left(nc(x^i,\sigma)-\mu_{nc(x,\sigma)}\right)^2}

    and

    .. math::

        nc(x^i,\sigma)=|\{x \in X_n, \|x-x^i\|<\sigma\}|-1

    :math:`\sigma` is the niche radius that is problem dependent and can be
    adjusted based on the distribution of the candidate solution in the space.
    :math:`\mu_{nc(x,\sigma)}` is the mean of the niche counts, :math:`nc`, calculated
    as :math:`\mu_{nc(x,\sigma)}=\frac{1}{|X_n|} \sum_{j=1}^{|X_n|} nc(x^j,\sigma)`.


    Parameters
    ----------
    nds
        Array with shape ``(n_solutions, n_objectives)`` containing the non-dominated
        solutions (NDS).  **All objectives are assumed to be minimised.**  If your
        problem involves maximisation, multiply the corresponding columns by ``-1``
        before calling.
    shared_dist
        Niche radius. Default is 0.01.

    Returns
    -------
        The uniform disttribution value.
    """

    if nds.ndim != 2:
        raise ValueError("nds.ndim should be 2")

    pf_size = len(nds)
    list_niches = numpy.zeros(pf_size)

    for i in range(0, pf_size):
        niche_count = 0

        for j in range(0, pf_size):
            if i != j:
                dist = numpy.linalg.norm(nds[i] - nds[j])
                niche_count += 1 if dist < shared_dist else 0

        list_niches[i] = niche_count

    niche_mean = numpy.mean(list_niches)

    # correction for the denominator of NaN
    if pf_size in [0, 1]:
        s_nc = 0.0
    else:
        s_nc = numpy.sqrt((numpy.sum((list_niches - niche_mean) ** 2)) / (pf_size - 1))

    return 1.0 / (1.0 + s_nc)


def _overall_spread(
    nds: numpy.typing.NDArray[numpy.float64],
    nadir_point: numpy.typing.ArrayLike,
    ideal_point: numpy.typing.ArrayLike,
) -> float:
    r"""Calculate overall pareto spread (OS).

    This method calculates diversity/spread based on how much extreme points are covered
    by Pareto Front approximation. Higher is better.

    This indicator assesses how well the points from the candidate set spreads towards
    the ideal of the optimal PF. OS is formulated as:

    .. math::
        OS(S,\mathcal{P})=\prod_{i=1}^{N}\left|\frac{\max\limits_{s \in S}s_i-\min\limits_{s \in S}s_i}{\max\limits_{p \in \mathcal{P}}p_{i}-\min\limits_{p \in \mathcal{P}}p_{i}}\right|

    Where the nominator and denominator are the absolute difference between the worst and
    best points for the candidate solution :math:`S` and Pareto optimal set :math:`\mathcal{P}`,
    respectively, on a particular metric dimension.


    Parameters
    ----------
    nds
        Array with shape ``(n_solutions, n_objectives)`` containing the non-dominated
        solutions (NDS).  **All objectives are assumed to be minimised.**  If your
        problem involves maximisation, multiply the corresponding columns by ``-1``
        before calling.
    nadir_point
        Nadir point with shape ``(n_objectives,)``. Must **not** dominate any point in
        ``nds``.
    ideal_point
        Reference point for the the ideal case where all objectives are fulfilled with
        shape ``(n_objectives,)``.  Must dominate any point in ``nds``.


    Returns
    -------
        The overall pareto spread value.

    Raises
    ------
    ValueError
        If ``nadir_point`` is not an 1-D array or does not match with ``nds``.
    """
    nadir_point_ = numpy.asarray(nadir_point, dtype=float)
    ideal_point_ = numpy.asarray(ideal_point, dtype=float)

    # validate
    if nds.size == 0:
        raise ValueError("nds must be non-empty.")

    if nds.ndim != 2:
        raise ValueError("nds.ndim should be 2")

    if nadir_point_.ndim != 1:
        raise ValueError(
            "nadir_point must be a 1‑D array‑like of shape (n_objectives,)."
        )

    if ideal_point_.ndim != 1:
        raise ValueError(
            "nadir_point must be a 1‑D array‑like of shape (n_objectives,)."
        )

    if nds.shape[1] != nadir_point_.shape[0]:
        raise ValueError(
            "The shape of nds.shape[1] and nadir_point.shape[0] should match."
        )

    if ideal_point_.shape[0] != nadir_point_.shape[0]:
        raise ValueError(
            "The shape of ideal_point.shape[0] and nadir_point.shape[0] should match."
        )

    numerator = nds.max(axis=0) - nds.min(axis=0)
    denominator = nadir_point_ - ideal_point_

    return numpy.prod(numerator / denominator)


def _average_spread(
    nds: numpy.typing.NDArray[numpy.float64],
    nadir_point: numpy.typing.ArrayLike,
    ideal_point: numpy.typing.ArrayLike,
) -> float:
    r"""Calculate average pareto spread (AS).

    This method calculates diversity/spread based on how much extreme points are covered
    by Pareto Front approximation. Higher is better.

    This indicator assesses how well the points from the candidate set spreads towards
    the ideal of the optimal PF. AS is formulated as:

    .. math::
        AS(S,\mathcal{P})=\frac{1}{N}\sum_{i=1}^{N}\left|\frac{\max\limits_{s \in S}s_i-\min\limits_{s \in S}s_i}{\max\limits_{p \in \mathcal{P}}p_{i}-\min\limits_{p \in \mathcal{P}}p_{i}}\right|

    Where the nominator and denominator are the absolute difference between the worst and
    best points for the candidate solution :math:`S` and Pareto optimal set :math:`\mathcal{P}`,
    respectively, on a particular metric dimension.


    Parameters
    ----------
    nds
        Array with shape ``(n_solutions, n_objectives)`` containing the non-dominated
        solutions (NDS).  **All objectives are assumed to be minimised.**  If your
        problem involves maximisation, multiply the corresponding columns by ``-1``
        before calling.
    nadir_point
        Nadir point with shape ``(n_objectives,)``. Must **not** dominate any point in
        ``nds``.
    ideal_point
        Reference point for the the ideal case where all objectives are fulfilled with
        shape ``(n_objectives,)``.  Must dominate any point in ``nds``.


    Returns
    -------
        The overall pareto spread value.

    Raises
    ------
    ValueError
        If ``nadir_point`` is not an 1-D array or does not match with ``nds``.
    """
    nadir_point_ = numpy.asarray(nadir_point, dtype=float)
    ideal_point_ = numpy.asarray(ideal_point, dtype=float)

    # validate
    if nds.size == 0:
        raise ValueError("nds must be non-empty.")

    if nds.ndim != 2:
        raise ValueError("nds.ndim should be 2")

    if nadir_point_.ndim != 1:
        raise ValueError(
            "nadir_point must be a 1‑D array‑like of shape (n_objectives,)."
        )

    if ideal_point_.ndim != 1:
        raise ValueError(
            "nadir_point must be a 1‑D array‑like of shape (n_objectives,)."
        )

    if nds.shape[1] != nadir_point_.shape[0]:
        raise ValueError(
            "The shape of nds.shape[1] and nadir_point.shape[0] should match."
        )

    if ideal_point_.shape[0] != nadir_point_.shape[0]:
        raise ValueError(
            "The shape of ideal_point.shape[0] and nadir_point.shape[0] should match."
        )

    numerator = nds.max(axis=0) - nds.min(axis=0)
    denominator = nadir_point_ - ideal_point_

    return numpy.average(numerator / denominator)


class Solutions(pydantic.BaseModel):
    """Data model representing system solutions (or utility/fairness trade-offs).

    Objects of this type carry information about two or more performance metrics
    (utility or fairness) for each operating mode (utility/fairness trade-off) of the
    analysed ML system.

    It is a dictionary where keys correspond to utility or fairness metrics calculated
    for the whole system, and values across different keys represent each the
    performance at a particular operating mode (utility/fairness trade-off) the system
    being analysed can potentially implement.
    """

    points: dict[str, list[typing.Annotated[float, pydantic.Field(ge=0.0, le=1.0)]]] = (
        pydantic.Field(
            ...,
            description="One or many lists of solution coordinates representing "
            "different operating mode of the system being analyzed.",
        )
    )

    metadata: dict[str, typing.Any] = pydantic.Field(
        ..., description="Additional information related to the points."
    )

    # --- Mapping API methods ---------------------------------------

    def __getitem__(self, key: str) -> list[float]:
        """Retrieve the list of floats associated with the given metric key.

        Parameters
        ----------
        key
            The metric name to look up.

        Returns
        -------
            The solution vector for the specified metric.

        Raises
        ------
        KeyError
            If the metric key is not present in the model.
        """
        return self.points[key]

    def __len__(self) -> int:
        """Return the number of metrics stored in the model.

        Returns
        -------
            The count of metric keys in the model.
        """
        return len(self.points)

    n_metrics = __len__

    def n_solutions(self) -> int:
        """Return the number of solutions stored in the model, across all metrics.

        Returns
        -------
            The number of solutions in the model, across all metrics.
        """
        return len(next(iter(self.points.values())))

    def items(self) -> typing.ItemsView[str, list[float]]:
        """Return a view of metric keys and their associated solution vectors.

        Returns
        -------
            A set-like view of (metric, vector) pairs.
        """
        return self.points.items()

    def keys(self) -> typing.KeysView[str]:
        """Return a view of the metric keys.

        Returns
        -------
            A set-like view of metric names.
        """
        return self.points.keys()

    def values(self) -> typing.ValuesView[list[float]]:
        """Return a view of all solution vectors.

        Returns
        -------
            A view of all metric solution vectors in the model.
        """
        return self.points.values()

    def __array__(self) -> numpy.typing.NDArray[numpy.float64]:
        """Allow numpy to convert this model directly into an (n_solutions, n_metrics) array.

        Each row is one solution; columns follow the insertion order of the metric keys.

        Parameters
        ----------
        dtype
            Optional dtype

        Returns
        -------
            A newly created and validated object.
        """
        # stack columns in the order of self.keys()
        return numpy.stack(
            [self.points[k] for k in self.keys()], axis=1, dtype=numpy.float64
        )

    @classmethod
    def fromarray(
        cls,
        points: numpy.typing.ArrayLike,
        metrics: typing.Sequence[str],
        thresholds: typing.Sequence[float],
        identifiers: typing.Sequence[int] | None = None,
        identifier_names: typing.Sequence[str] | None = None,
    ) -> typing.Self:
        """Create a new instance from an array and names of metrics.

        Parameters
        ----------
        points
            2-D array-like object with floating-point numbers organized as
            ``(n_solutions, n_metrics)``.
        metrics
            A set of strictly **valid and supported** metrics, each representing the
            columns of the input points array.
        thresholds
            A list of tresholds that were used to compute the points.
        identifiers
            Optional list of identifiers, indicating which model each point corresponds to.
            Each identifier indexes into identifier_names.
            If not specified, all points will have the same identifier.
        identifier_names
            Optoinal list of identifier names.

        Returns
        -------
            A newly created and validated object.

        Raises
        ------
        AssertionError
            If the number of columns on the input array-like object is different than
            the number of listed metrics.
        """
        points_arr = numpy.asarray(points, dtype=numpy.float64)
        assert points_arr.shape[1] == len(metrics)

        if identifiers is not None and identifier_names is None:
            logger.warning(
                "No names corresponding to the identifiers specified. Default ones will be generated."
            )

            identifier_names = [
                f"model-{i}" for i in range(min(identifiers), max(identifiers) + 1)
            ]

        if identifiers is None:
            logger.warning(
                "No identifiers specified when calling Solutions.fromarray(). All points will be set to the same identifier."
            )
            identifier_names = ["model-0"]
            identifiers = numpy.full(points_arr.shape[0], 0)

        assert len(identifiers) == points_arr.shape[0]

        return cls.model_validate(
            {
                "points": dict(zip(metrics, points_arr.T)),
                "metadata": {
                    "identifier-names": identifier_names,
                    "identifiers": identifiers,
                    "thresholds": thresholds,
                },
            }
        )

    @classmethod
    def load(cls, source: pathlib.Path | str | typing.TextIO) -> typing.Self:
        """Validate and load a JSON file into a solution data object.

        This function is intended to validate and load the input in JSON format. It opens
        the given file path, parses its JSON content, and validates it against the defined
        model.

        Parameters
        ----------
        source
            Source input where to read JSON from.

        Returns
        -------
            Parsed and validated content as a :py:class:`Solutions` instance.

        Raises
        ------
        pydantic_core.ValidationError
            If the file contains invalid data.
        """

        if isinstance(source, pathlib.Path | str):
            path = pathlib.Path(source)
            return cls.model_validate_json(
                path.read_text(), context={"base_dir": path.parent}
            )

        else:  # noqa: RET505
            return cls.model_validate_json(source.read())

    def save(self, dest: pathlib.Path | str | typing.TextIO, **args) -> None:
        """Save contents to an external file.

        Parameters
        ----------
        dest
            Destination where to save the contents. If not a path or str, then assumed
            to have a ``write`` method accepting strings.
        args
            Parameters further passed down to
            :py:func:`pydantic.BaseModel.model_dump_json`.
        """

        if isinstance(dest, pathlib.Path | str):
            with pathlib.Path(dest).open("w", encoding="utf-8") as f:
                f.write(self.model_dump_json(**args))

        else:
            dest.write(self.model_dump_json(**args))

    @pydantic.model_validator(mode="after")
    def check_metrics_validity(self) -> typing.Self:
        """Ensure all metrics are valid."""

        invalid = []
        parsed: list[
            metrics.UtilityMetricsType
            | tuple[metrics.FairnessMetricsType, str]
            | tuple[metrics.MinMaxFairnessMetricsType, metrics.UtilityMetricsType, str]
        ] = []
        for name in self.points.keys():
            try:
                parsed.append(metrics.parse_metric(name))
            except ValueError:
                invalid.append(f"`{name}`")

        if invalid:
            raise ValueError(f"invalid metric names: {', '.join(invalid)}")

        # cannot have utility metrics relying on statistics from a single utility group
        # (positives or negatives), as those will generate misleading results -- see
        # discussion at: https://gitlab.idiap.ch/medai/software/fairical/-/merge_requests/4
        utility_metrics: list[metrics.UtilityMetricsType] = [
            typing.cast(metrics.UtilityMetricsType, k)
            for k in parsed
            if k in typing.get_args(metrics.UtilityMetricsType)
        ]
        if len(utility_metrics) == 1:
            if utility_metrics[0] in ("fpr", "tpr", "fnr", "tnr", "rec"):
                raise ValueError(
                    f"cannot use `{utility_metrics[0]}` as the sole utility metrics "
                    f"for analysis - it generates biased results at extreme thresholds"
                )
        elif len(utility_metrics) == 2:
            forbidden_groups = [
                set(k) for k in [("fpr", "tnr"), ("fnr", "tpr"), ("fnr", "rec")]
            ]
            if set(utility_metrics) in forbidden_groups:
                raise ValueError(
                    f"cannot use the pair `{utility_metrics}` as the sole utility "
                    f"metrics for analysis as both measures belong to the same "
                    f"group -- it generates biased results at extreme thresholds"
                )

        return self

    @pydantic.model_validator(mode="after")
    def check_identifiers_specified(self) -> typing.Self:
        """Ensure solutions contain identifiers."""
        if "identifiers" not in self.metadata:
            raise ValueError("No identifiers found in solutions")
        return self

    @pydantic.model_validator(mode="after")
    def check_consistent_lengths(self) -> typing.Self:
        """Ensure all solution lists have the same length."""

        expected = len(next(iter(self.points.values())))

        for key, value in self.points.items():
            if len(value) != expected:
                raise ValueError(f"solutions[{key}] length {len(value)} != {expected}")

        identifiers_length = len(self.metadata["identifiers"])
        if identifiers_length != expected:
            raise ValueError(f"identifiers length {identifiers_length} != {expected}")

        return self

    def filter_metadata_by_indices(self, indices):
        """Filter metadata from provided indices.

        Parameters
        ----------
        indices
            The indices used for filtering.

        Returns
        -------
            Filtered metadata.
        """
        metadata = self.metadata.copy()

        if len(indices) != 0:
            # identifiers
            identifiers = numpy.asarray(metadata["identifiers"])

            filtered_identifiers = identifiers[numpy.sort(indices)].tolist()
            index_to_name = {
                idx: metadata["identifier-names"][idx]
                for idx in set(filtered_identifiers)
            }

            unique_identifiers = set(filtered_identifiers)
            metadata["identifier-names"] = [
                metadata["identifier-names"][identifier]
                for identifier in unique_identifiers
            ]

            # since names have been removed, identifiers must be re-indexed
            name_to_new_index = {
                name: idx for idx, name in enumerate(metadata["identifier-names"])
            }

            index_to_new_index = {
                k: name_to_new_index[v] for k, v in index_to_name.items()
            }

            metadata["identifiers"] = [
                index_to_new_index[i] for i in filtered_identifiers
            ]

            # thresholds
            metadata["thresholds"] = [
                metadata["thresholds"][i] for i in sorted(indices)
            ]
        else:
            metadata["identifiers"] = []
            metadata["identifier-names"] = []
            metadata["thresholds"] = []

        return metadata

    def deduplicate(self, eps: float = 1e-6) -> typing.Self:
        """Filter solutions to remove duplicates within a certain epsilon.

        Remove points in these solutions that lie within ``eps`` of another by
        clustering with :py:class:`sklearn.cluster.DBSCAN` (min_samples=1) and keeping
        the first point in each cluster.

        Parameters
        ----------
        eps
            Maximum distance between points in the same cluster.

        Returns
        -------
            Filtered solutions without duplicates, as a new object.
        """
        dims = list(self.points.keys())
        metadata = self.metadata.copy()

        points = numpy.vstack([self.points[d] for d in dims]).T

        labels = sklearn.cluster.DBSCAN(
            eps=eps, min_samples=1, metric="euclidean"
        ).fit_predict(points)

        # pick the first index for each label
        _, first_idx = numpy.unique(labels, return_index=True)
        filtered = points[numpy.sort(first_idx)]

        logger.info(
            f"Deduplication reduced solutions {self.n_solutions()} -> {len(filtered)}"
        )

        filtered = {dim: filtered[:, idx].tolist() for idx, dim in enumerate(dims)}
        metadata = self.filter_metadata_by_indices(first_idx)

        return self.model_validate({"points": filtered, "metadata": metadata})

    def non_dominated_solutions(self) -> tuple[typing.Self, typing.Self]:
        """Filter solutions from system that are non-dominated.

        This is a thin wrapper around :py:class:`pymoo.util.nds.NonDominatedSorting` that
        extracts the *rank‑0* solutions (those that are not dominated by any other).

        **Definition**: A point p is dominated only if one single competitor is no worse in
        every objective and strictly better in at least one.

        Parameters
        ----------
        solutions
            All solutions available in the current system.

        Returns
        -------
            A tuple containing non-dominated and dominated solutions respectively. By
            definition, the sets are guaranteed to not overlap.
        """

        # calculation of non-dominated solutions requires all axes to be set as one is
        # minimizing on every direction.
        points = numpy.asarray(self)
        for k, metric in enumerate(self.points.keys()):
            if not metrics.should_minimize(metric):
                points[:, k] *= -1

        algo = pymoo.util.nds.non_dominated_sorting.NonDominatedSorting()
        # nds_idx = set([int(p) for p in set(algo.do(points, only_non_dominated_front=True))])
        nds_idx = typing.cast(
            set[int], set(algo.do(points, only_non_dominated_front=True))
        )

        ds_idx = set(range(self.n_solutions())) - nds_idx

        def filter_by_indices(points, indices):
            return {k: [points[k][i] for i in indices] for k in points.keys()}

        def get_solutions(indices):
            points = filter_by_indices(self.points, sorted(indices))
            metadata = self.filter_metadata_by_indices(sorted(indices))

            return self.model_validate({"points": points, "metadata": metadata})

        nds = self.model_validate(get_solutions(list(nds_idx)))
        ds = self.model_validate(get_solutions(list(ds_idx)))

        logger.info(
            f"Non-dominated/Dominated solutions {nds.n_solutions()}/{ds.n_solutions()}"
        )

        return nds, ds

    def indicators(self) -> dict[utils.IndicatorType, float]:
        r"""Assess utility-fairness trade-off systems based on characteristics of the
        estimated Pareto front.

        This method evaluates trade-off between utiltiy and fairness of adjustable systems
        by using Multi-Objective based performance indicators. It first estimates the
        set of non-dominated solutions.

        Returns
        -------
            A dictionary that characterizes the (estimated Pareto) front composed of
            non-dominated solutions in ``nds``. The dictionary contains the following keys:

            * ``hv``: The hypervolume of the front.

              Higher is better. This indicator evaluates how the solution set covers the
              metric space in terms of diversity and proximity to the ideal. HV is
              formulated as:

              .. math::
                  HV(S) = VOL\left(\bigcup_{\substack{x \in S \\ x \prec r}} \prod_{i=1}^{N}[x^{i},r^{i}]\right)

              Where :math:`x` is the solution set and :math:`r` is the Nadir point

            * ``ud``: Uniformity of the distribution of ``nds`` points on the front.

              This indicator evaluates how uniform the solution set is spanned in the
              metric space based on an upper-bound distance, :math:`\sigma`. UD is
              formulated as:

              .. math::

                 UD(S,\sigma)=\frac{1}{1+D_{nc}(S, \sigma)}

              Where

              .. math::

                  D_{nc}(S,\sigma)=\sqrt{\frac{1}{|X_n|-1} \sum_{i=1}^{|X_n|} \left(nc(x^i,\sigma)-\mu_{nc(x,\sigma)}\right)^2}

              and

              .. math::

                  nc(x^i,\sigma)=|\{x \in X_n, \|x-x^i\|<\sigma\}|-1

              :math:`\sigma` is the niche radius that is problem dependent and can be
              adjusted based on the distribution of the candidate solution in the space.
              :math:`\mu_{nc(x,\sigma)}` is the mean of the niche counts, :math:`nc`,
              calculated as :math:`\mu_{nc(x,\sigma)}=\frac{1}{|X_n|} \sum_{j=1}^{|X_n|}
              nc(x^j,\sigma)`.

            * ``os``: Overall spread of ``nds`` points with respect to extremities of the front.

              This indicator assesses how well the points from the candidate set spreads
              towards the ideal of the optimal PF. OS is formulated as:

              .. math::
                  OS(S,\mathcal{P})=\prod_{i=1}^{N}\left|\frac{\max\limits_{s \in S}s_i-\min\limits_{s \in S}s_i}{\max\limits_{p \in \mathcal{P}}p_{i}-\min\limits_{p \in \mathcal{P}}p_{i}}\right|

              Where the nominator and denominator are the absolute difference between
              the worst and best points for the candidate solution :math:`S` and Pareto
              optimal set :math:`\mathcal{P}`, respectively.

            * ``as``: Average spread of ``nds`` points with respect to extremities of the front.

              This indicator assesses how well the points from the candidate set spreads
              towards the ideal of the optimal PF. AS is formulated as:

              .. math::
                  AS(S,\mathcal{P})=\frac{1}{N}\sum_{i=1}^{N}\left|\frac{\max\limits_{s \in S}s_i-\min\limits_{s \in S}s_i}{\max\limits_{p \in \mathcal{P}}p_{i}-\min\limits_{p \in \mathcal{P}}p_{i}}\right|

              Where the nominator and denominator are the absolute difference between
              the worst and best points for the candidate solution :math:`S` and Pareto
              optimal set :math:`\mathcal{P}`, respectively.

            * ``onvg``: Overall Nondominated Vector Generation (ONVG) in the front
              (``nds``).

              Higher is better. This indicator evaluates how many optimal solutions are
              generated by the system. ONVG is formulated as:

              .. math::
                  ONVG(S) = |X_n|

              Where :math:`|.|` is the cardinality of the candidate solution set in the metric space.

            * ``onvgr``: Ratio between number of solutions in ``nds`` and ``nds + ds``.

              Higher is better. This indicator assesses the proportion of optimal solutions
              generated by the system. ONVGR is formulated as:

              .. math::
                  ONVGR(S) = \left|\frac{X_n}{S}\right|

              Where :math:`|.|` is the ratio of the optimality.
        """

        assert len(self.points) >= 2, (
            "cannot extract indicators on solutions with a single dimension"
        )

        nds, ds = self.non_dominated_solutions()

        # calculate indicators
        retval: dict[utils.IndicatorType, float] = {}
        nds_arr = numpy.asarray(nds)

        # use the complement to 1.0, to everything we need to maximize so we can keep
        # using (1, 1, ...) as a fixed nadir point, and (0, 0, ...) for the ideal point.
        for k, metric in enumerate(self.points.keys()):
            if not metrics.should_minimize(metric):
                nds_arr[:, k] *= -1
                nds_arr[:, k] += 1

        nadir_point = len(self) * (1.0,)  # dominated by any point in nds
        ideal_point = len(self) * (0.0,)  # dominates all points in nds

        retval["hv"] = _hypervolume(nds_arr, nadir_point)
        retval["ud"] = _uniform_distribution(nds_arr, shared_dist=0.01)
        retval["os"] = _overall_spread(nds_arr, nadir_point, ideal_point)
        retval["as"] = _average_spread(nds_arr, nadir_point, ideal_point)
        retval["onvg"] = float(nds.n_solutions())
        retval["onvgr"] = float(
            nds.n_solutions() / (nds.n_solutions() + ds.n_solutions())
        )

        return retval

    def tabulate(self) -> str:
        """Generate a table containing the given solutions.

        Each table row contains the values for each metric,
        the threshold used to compute them, and the corresponding identifier name.

        Returns
        -------
            The generated table as a string.

        """
        metrics = ", ".join(self.points.keys())
        metrics = f"({metrics})"

        points = list(zip(*self.points.values()))

        front_data = {
            metrics: points,
            "thresholds": self.metadata["thresholds"],
            "identifiers": [
                self.metadata["identifier-names"][idx]
                for idx in self.metadata["identifiers"]
            ],
        }

        return tabulate.tabulate(front_data, headers="keys")



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib

import matplotlib
import pytest

matplotlib.use("agg")


@pytest.fixture
def datadir(request) -> pathlib.Path:
    """Return the directory in which the test is sitting. Check the pytest
    documentation for more information.

    Parameters
    ----------
    request
        Information of the requesting test function.

    Returns
    -------
    pathlib.Path
        The directory in which the test is sitting.
    """

    return pathlib.Path(request.module.__file__).parents[0]



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy
import pytest

from fairical.solutions import _uniform_distribution as ud


def test_ud_perfectly_uniform() -> None:
    """Three points that are evenly spaced inside radius."""
    pts = numpy.array([[0.0, 0.0], [0.005, 0.0], [0.01, 0.0]])
    result = ud(pts, shared_dist=0.01)
    assert 0 < result <= 1
    assert numpy.isclose(result, 0.63, 0.01)


def test_ud_single_point() -> None:
    """Single point."""
    pts = numpy.array([[0.0, 0.0]])
    assert numpy.isclose(ud(pts), 1.0)  # No deviation with 1 point


def test_ud_two_points_within_radius() -> None:
    """Two points inside radius."""
    pts = numpy.array([[0.0, 0.0], [0.005, 0.0]])
    assert numpy.isclose(
        ud(pts, shared_dist=0.01), 1.0
    )  # Niche counts are same (1 each), so std = 0


def test_ud_two_points_outside_radius() -> None:
    """Two points outside radius."""
    pts = numpy.array([[0.0, 0.0], [1.0, 1.0]])
    assert numpy.isclose(
        ud(pts, shared_dist=0.01), 1.0
    )  # Both have 0 neighbors, std = 0


def test_ud_non_uniform_distribution() -> None:
    """Three points as non uniformly distributed."""
    pts = numpy.array([[0, 0], [0.001, 0.0], [1, 1]])
    assert 0 < ud(pts, shared_dist=0.01) < 1  # Non-uniform distribution


def test_ud_empty_input() -> None:
    """Empty input."""
    pts = numpy.empty((0, 2))
    assert ud(pts) == 1.0  # The worst case ending up with full non-uniformity.


@pytest.mark.parametrize(
    "pts",
    [
        (numpy.array([1.0, 2.0, 3.0])),  # 1-D
        (numpy.ones((2, 2, 2))),  # 3-D
    ],
)
def test_ud_invalid_shape_raises(pts: numpy.ndarray) -> None:
    """Invalid input in 1-D and 3-D."""
    with pytest.raises(ValueError):
        ud(pts)



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Helpers to evaluate scikit-learn metrics at arbitrary thresholds."""

import typing

import fairlearn.metrics
import numpy
import sklearn.metrics

UtilityMetricsType: typing.TypeAlias = typing.Literal[
    "fpr",  # False Positive Rate (minimize, range: [0, 1])
    "tpr",  # True Positive Rate (maximize, range: [0, 1])
    "tnr",  # True Negative Rate (minimize, range: [0, 1])
    "fnr",  # False Negative Rate (minimize, range: [0, 1])
    "roc_auc",  # Area Under the Curve for Receiver Operating Characteristic (maximize, range: [0, 1])
    "prec",  # Precision (maximize, range: [0, 1])
    "rec",  # Recall (maximize, range: [0, 1])
    "avg_prec",  # Average precision for Precision Recall Curve (maximize, range: [0, 1])
    "f1",  # F1 Score (maximize, range: [0, 1])
    "acc",  # Accuracy (maximize, range: [0, 1])
    "bal_acc",  # Balanced Accuracy (maximize, range: [0, 1])
]
"""Supported utility metrics type for pareto front estimates."""

FairnessMetricsType: typing.TypeAlias = typing.Literal[
    "dpd",  # Demographic Parity Difference (minimize, range: [0, 1])
    "dpr",  # Demographic Parity Ratio (maximize, range: [0, 1])
    "eod",  # Equalized Odds Difference (minimize, range: [0, 1])
    "eor",  # Equalized Odds Ratio (maximize, range: [0, 1])
]
"""Supported fairness metrics type for pareto front estimates."""

MinMaxFairnessMetricsType: typing.TypeAlias = typing.Literal[
    "minmaxd",  # Min-Max (absolute) difference: max(util) - min(util)
    "minmaxr",  # Min-Max ratio: min(util) / max(util)
]
"""Supported min-max fairness metrics type for pareto front estimates."""


def parse_metric(
    name: str,
) -> (
    UtilityMetricsType
    | tuple[FairnessMetricsType, str]
    | tuple[MinMaxFairnessMetricsType, UtilityMetricsType, str]
):
    """Parse and validate a string supposed to carry a metric name.

    Valid metric names are the ones listed in :py:type:`UtilityMetricsType`,
    :py:type:`FairnessMetricsType` (followed by a "+<attr>"), or
    :py:type:`MinMaxFairnessMetricsType` (followed by a "+<util>+<attr>"), where
    ``<attr>`` corresponds to the protected attribute being measured by the fairness
    metric, and ``<util>`` corresponds to the :py:type:`UtilityMetricsType` to be used
    to measure min-max fairness difference or ratios.

    Parameters
    ----------
    name
        The string to be validated.

    Returns
    -------
        The parsed metric.

    Raises
    ------
    ValueError
        If the metric expressed in ``name`` is invalid.
    """

    parts = name.split("+", 2)
    if parts[0] in typing.get_args(UtilityMetricsType):
        return typing.cast(UtilityMetricsType, parts[0])
    if parts[0] in typing.get_args(FairnessMetricsType):
        if len(parts) != 2:
            raise ValueError(
                f"fairness metric should be set like `{parts[0]}+<attr>` "
                f"(`{name}` is invalid)"
            )
        return (typing.cast(FairnessMetricsType, parts[0]), parts[1])
    if parts[0] in typing.get_args(MinMaxFairnessMetricsType):
        if len(parts) != 3:
            raise ValueError(
                f"min-max fairness metric should be set like `{parts[0]}+<util>+<attr>` "
                f"(`{name}` is invalid)"
            )
        if parts[1] not in typing.get_args(UtilityMetricsType):
            raise ValueError(f"invalid utility metric name `{parts[1]}` at `{name}`")
        return (
            typing.cast(MinMaxFairnessMetricsType, parts[0]),
            typing.cast(UtilityMetricsType, parts[1]),
            parts[2],
        )

    raise ValueError(f"Invalid metric specification: `{name!r}`")


def should_minimize(metric: str) -> bool:
    """For a given metric, tells if it should be minimized or maximized.

    Currently, "fpr", in the utility side, "minmaxd" on the min-max fairness metrics, or
    any other fairness metric should be minimized. All others should be maximized.

    Parameters
    ----------
    metric
        Metric name.

    Returns
    -------
        ``True``, if the metric should be minimized (instead of maximized).  ``False``
        otherwise.

    Raises
    ------
    ValueError
        If the metric is invalid.
    """
    pm = parse_metric(metric)
    return (
        pm in ("fpr", "fnr")
        or (isinstance(pm, tuple) and len(pm) == 2 and pm[0] in ("dpd", "eod"))
        or (isinstance(pm, tuple) and len(pm) == 3 and pm[0] == "minmaxd")
    )


def supported_metrics() -> list[str]:
    """Generate a comma-separated list of supported metrics.

    Returns
    -------
        A comma-separated list of supported metrics.
    """

    utility = typing.get_args(UtilityMetricsType)
    fairness = [f"{k}+<attr>" for k in typing.get_args(FairnessMetricsType)]
    minmax = [f"{k}+<util>+<attr>" for k in typing.get_args(MinMaxFairnessMetricsType)]

    return list(utility) + fairness + minmax


def calculate_metric(
    metric: str,
    y_true: typing.Sequence[int],
    y_score: typing.Sequence[float],
    thresholds: typing.Sequence[float],
    sensitive_attributes: typing.Mapping[str, typing.Sequence[int | str]] | None = None,
) -> list[float]:
    """Entry-point function to calculate arbitrary (supported) metrics.

    This function works as an entry-point to the metric calculation submodule. It can
    calculate arbirary (supported) metrics provided input information for a system,
    consisting of ground-truth, scores, thresholds and (optionally) sensitive features.

    Parameters
    ----------
    metric
        The metric to calculate.
    y_true
        True binary labels (0 or 1).
    y_score
        Predicted continuous scores or probabilities.
    thresholds
        Threshold values at which to binarize ``y_score`` (:math:`score >= threshold`
        implies sample is classified as positive).
    sensitive_attributes
        Group membership for each sample, according to protected attribute. Only
        required if ``metric`` is a fairness metric. Each entry in the input dictionary
        should match the order of samples in ``y_true`` and ``y_score``. When ``metric``
        refers to a particular sensitive attribute, it should be a key in this
        dictionary.

    Returns
    -------
        The metric over all considered thresholds.

    Raises
    ------
    ValueError
        In case of unknown metrics.
    """
    parsed_metric = parse_metric(metric)

    if parsed_metric in typing.get_args(UtilityMetricsType):
        # simple closure to avoid repeatitive for loops with the same config
        def _for_all_t(f):
            y_score_arr = numpy.asarray(y_score, dtype=float)
            return numpy.nan_to_num(
                [f(y_true, y_score_arr >= t) for t in thresholds]
            ).tolist()

        match metric:
            case "fpr":
                return _for_all_t(fairlearn.metrics.false_positive_rate)

            case "tpr":
                return _for_all_t(fairlearn.metrics.true_positive_rate)

            case "fnr":
                return _for_all_t(fairlearn.metrics.false_negative_rate)

            case "tnr":
                return _for_all_t(fairlearn.metrics.true_negative_rate)

            case "acc":
                return _for_all_t(sklearn.metrics.accuracy_score)

            case "bal_acc":
                return _for_all_t(sklearn.metrics.balanced_accuracy_score)

            case "prec":
                return _for_all_t(sklearn.metrics.precision_score)

            case "rec":
                return _for_all_t(sklearn.metrics.recall_score)

            case "f1":
                return _for_all_t(sklearn.metrics.f1_score)

            case "roc_auc":
                # there is only 1 roc-auc, where there are multiple fpr or tpr points --
                # so we repeat the roc-auc as many times as there are thresholds to keep
                # consistence.
                val = float(sklearn.metrics.roc_auc_score(y_true, y_score))
                return len(thresholds) * [val]

            case "avg_prec":
                # there is only 1 average precision, where there are multiple fpr or tpr
                # points -- so we repeat the average precision as many times as there are
                # thresholds to keep consistence.
                val = float(sklearn.metrics.average_precision_score(y_true, y_score))
                return len(thresholds) * [val]

    elif (
        isinstance(parsed_metric, tuple)
        and len(parsed_metric) == 2
        and parsed_metric[0] in typing.get_args(FairnessMetricsType)
    ):
        assert sensitive_attributes is not None
        assert parsed_metric[1] in sensitive_attributes

        # simple closure to avoid repeatitive for loops with the same config
        def _for_all_t(f):
            y_score_arr = numpy.asarray(y_score, dtype=float)
            return numpy.nan_to_num(
                [
                    f(
                        y_true,
                        y_score_arr >= t,
                        sensitive_features=sensitive_attributes[parsed_metric[1]],
                    )
                    for t in thresholds
                ]
            ).tolist()

        match parsed_metric[0]:
            case "dpd":
                return _for_all_t(fairlearn.metrics.demographic_parity_difference)

            case "dpr":
                return _for_all_t(fairlearn.metrics.demographic_parity_ratio)

            case "eod":
                return _for_all_t(fairlearn.metrics.equalized_odds_difference)

            case "eor":
                return _for_all_t(fairlearn.metrics.equalized_odds_ratio)

    elif (
        isinstance(parsed_metric, tuple)
        and len(parsed_metric) == 3
        and parsed_metric[0] in typing.get_args(MinMaxFairnessMetricsType)
    ):
        assert sensitive_attributes is not None
        assert parsed_metric[2] in sensitive_attributes

        # simple closures to avoid repeatitive for loops with the same config
        def _for_all_t(f):
            y_score_arr = numpy.asarray(y_score, dtype=float)
            return numpy.nan_to_num(
                [
                    f(
                        y_true,
                        y_score_arr >= t,
                        sensitive_features=sensitive_attributes[parsed_metric[2]],
                    )
                    for t in thresholds
                ]
            ).tolist()

        def _make_derived(f):
            return fairlearn.metrics.make_derived_metric(
                metric=f,
                transform="difference" if parsed_metric[0] == "minmaxd" else "ratio",
            )

        match parsed_metric[1]:
            case "fpr":
                return _for_all_t(_make_derived(fairlearn.metrics.false_positive_rate))

            case "tpr":
                return _for_all_t(_make_derived(fairlearn.metrics.true_positive_rate))

            case "fnr":
                return _for_all_t(_make_derived(fairlearn.metrics.false_negative_rate))

            case "tnr":
                return _for_all_t(_make_derived(fairlearn.metrics.true_negative_rate))

            case "acc":
                return _for_all_t(_make_derived(sklearn.metrics.accuracy_score))

            case "bal_acc":
                return _for_all_t(
                    _make_derived(sklearn.metrics.balanced_accuracy_score)
                )

            case "prec":
                return _for_all_t(_make_derived(sklearn.metrics.precision_score))

            case "rec":
                return _for_all_t(_make_derived(sklearn.metrics.recall_score))

            case "f1":
                return _for_all_t(_make_derived(sklearn.metrics.f1_score))

            case "roc_auc":
                # there is only 1 roc-auc, where there are multiple fpr or tpr points --
                # so we repeat the roc-auc as many times as there are thresholds to keep
                # consistence.
                val = float(
                    _make_derived(sklearn.metrics.roc_auc_score)(
                        y_true,
                        y_score,
                        sensitive_features=sensitive_attributes[parsed_metric[2]],
                    )
                )
                return len(thresholds) * [val]

            case "avg_prec":
                # there is only 1 average precision, where there are multiple fpr or tpr
                # points -- so we repeat the average precision as many times as there are
                # thresholds to keep consistence.
                val = float(
                    _make_derived(sklearn.metrics.average_precision_score)(
                        y_true,
                        y_score,
                        sensitive_features=sensitive_attributes[parsed_metric[2]],
                    )
                )
                return len(thresholds) * [val]

    # this should not occur, as metric is parsed from start
    raise ValueError(f"Invalid metric specification: `{metric!r}`")



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib
import typing

import numpy
import pytest

from fairical.scores import Scores
from fairical.solutions import Solutions
from fairical.utils import IndicatorType


@pytest.mark.parametrize(
    "system",
    [
        ("empirical"),
        ("sample"),
    ],
    ids=lambda x: str(x),  # just changes how pytest prints it
)
def test_scores_case(
    datadir: pathlib.Path,
    system: str,
) -> None:
    thresholds = typing.cast(list[float], list(numpy.linspace(0, 1, 5, dtype=float)))

    data = Scores.load(datadir / "data" / system / "system_1.json")
    solutions = data.solutions_a_posteriori(
        ["eod+race", "minmaxr+roc_auc+gender", "acc"], thresholds=thresholds
    ).deduplicate()

    indicators = solutions.indicators()

    assert isinstance(indicators, dict)
    assert len(indicators.keys()) == 6

    for ind in [
        # "onvg", >> 1
        "onvgr",
        # "relative-onvg",
        "ud",
        # "complement-ud",
        "os",
        "as",
        "hv",
        # "area",
    ]:
        assert ind in indicators
        assert 0.0 <= indicators[typing.cast(IndicatorType, ind)] <= 1.0
    assert indicators["onvg"] >= 1

    expected = len(next(iter(solutions.points.values())))
    assert len(solutions.metadata["thresholds"]) == expected


@pytest.mark.parametrize(
    "system",
    [
        ("uc-1"),
        ("uc-2"),
        ("uc-3"),
    ],
    ids=lambda x: str(x),  # just changes how pytest prints it
)
def test_solutions_case(
    datadir: pathlib.Path,
    system: str,
) -> None:
    data = Solutions.load(datadir / "data" / system / "system_1.json")
    indicators = data.indicators()

    assert isinstance(indicators, dict)
    assert len(indicators.keys()) == 6

    for ind in [
        # "onvg", >> 1
        "onvgr",
        # "relative-onvg",
        "ud",
        # "complement-ud",
        "os",
        "as",
        "hv",
        # "area",
    ]:
        assert ind in indicators
        assert 0.0 <= indicators[typing.cast(IndicatorType, ind)] <= 1.0
    assert indicators["onvg"] >= 1



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Solver CLI for demographically fair machine learning systems.

This module defines the ``solve`` command, which loads one or more system score output
files in JSON format, validates them against a predefined data model, and computes
fairness-related indicators such as accuracy and equalized odds. It outputs JSON files
containing solutions (or operating-modes) following the chosen metrics space, and
thresholds.

Input files must conform to a structured data model including scores, ground-truth
labels, and protected attributes. The CLI provides options to customize metrics, and
select thresholds.
"""

import logging
import pathlib
import typing

logger = logging.getLogger(__name__.split(".", 1)[0])

import click

from ..metrics import supported_metrics
from ..scores import Scores
from ..solutions import Solutions
from . import utils


@click.command(
    epilog="""Examples:

\b
  1. Compute and store solutions from two systems on the current directory:

     .. code:: sh

        fairical solve scores/system1.json scores/system2.json

  2. Compute and store a-priori solutions for system1, using the thresholds
     from the non-dominated solutions of system2:

     .. code:: sh

        fairical solve scores/system1.json@solutions/system2_nds.json

""",
)
@click.argument(
    "system",
    type=str,
    required=True,
    nargs=-1,
    callback=utils.validate_scores_json,
)
@click.option(
    "-d/-D",
    "--deduplicate/--no-deduplicate",
    default=True,
    show_default=True,
    help="If we should prune (deduplicate) similar solutions.",
)
@click.option(
    "--ds/--nds",
    "dominated",
    default=None,
    flag_value=True,
    help="This flag only concerns *a priori* analysis. If --ds is set, "
    "then only applies only dominated solutions to input score set, if "
    "--nds, only applies non-dominated solutions. If neither of these "
    "flags is set, then apply all solutions from the input solution set "
    "during analysis.",
)
@click.option(
    "-o",
    "--output-path",
    type=click.Path(
        dir_okay=True, file_okay=False, writable=True, path_type=pathlib.Path
    ),
    required=True,
    default=pathlib.Path(),
    show_default=True,
    help="Directory where to store system solutions.",
)
@click.option(
    "-m",
    "--metric",
    type=str,
    multiple=True,
    help=f"Metric to consider for solution evaluation. Values are considered in the "
    f"order they are provided forming the axes of a potentially multi-dimensional "
    f"surface where dominated and non-dominated solutions are evaluated. Valid values "
    f"are: {', '.join(supported_metrics())}, where ``<attr>`` refers to attributes in "
    f"the provided sensitive attribute dictionary, and ``<util>`` refers to any "
    f"supported utility metrics.  This option is ignored for *a priori* analysis.",
    default=["acc", "eod+gender"],
    callback=utils.validate_metrics,
    show_default=True,
)
@click.option(
    "-T",
    "--thresholds",
    type=click.IntRange(2),
    help="If set, then run for a fixed number of equally-spaced thresholds in the "
    "interval [0.0, 1.0].  Here, you only specify the number of thresholds to use, "
    "with the minimum being 2. If you want explicitly set the thresholds, you should "
    "use the API directly. If unset, then let scikit-learn calculate the number of "
    "sensible thresholds based the scores and changes on the ROC or PR-curve "
    "depending on utility metrics chosen.",
    default=None,
    show_default=True,
    callback=utils.validate_thresholds,
)
@utils.verbosity_option(logger)
def solve(
    system: list[
        tuple[pathlib.Path, Scores]
        | tuple[tuple[pathlib.Path, Scores], tuple[pathlib.Path, Solutions]]
    ],
    deduplicate: bool,
    dominated: bool | None,
    output_path: pathlib.Path,
    metric: list[str],
    thresholds: list[float] | None,
    verbose: int,
) -> None:  # numpydoc ignore=PR01
    """Find operating-modes (a.k.a. solutions) for ML systems.

    This module defines the ``solve`` command, which loads one or more system score
    output files in JSON format, validates them against a predefined data model, and
    computes fairness-related indicators such as accuracy and equalized odds. It outputs
    JSON files containing solutions (or operating-modes) following the chosen metrics
    space, and thresholds.

    Input JSON files must be formatted as described in the documentation.
    """
    # 0. validate that the selected protected attributes do exist inside JSON system
    # TODO: validate metrics against user loaded systems

    # 2. for each system provided as input, get solutions
    for k in system:
        if isinstance(k[0], pathlib.Path):
            # solutions chosen a posteriori
            score_fn, scores = typing.cast(tuple[pathlib.Path, Scores], k)

            message = "Calculating solutions *a posteriori* for system "
            message += f"`{str(score_fn)}`, using metrics "
            message += f"`{', '.join(metric)}`..."
            logger.info(message)

            solutions = scores.solutions_a_posteriori(metric, thresholds=thresholds)

            output_filename = (output_path / k[0].stem).with_suffix(".json")

        else:
            # solutions chosen a priori
            (score_fn, scores), (solution_fn, solutions_a_priori) = typing.cast(
                tuple[tuple[pathlib.Path, Scores], tuple[pathlib.Path, Solutions]], k
            )

            metric = list(solutions_a_priori.points.keys())

            message = f"Calculating solutions *a priori* for system `{str(score_fn)}`"
            message += f" based on solutions from `{str(solution_fn)}`, using metrics "
            message += f"`{', '.join(metric)}`..."
            logger.info(message)

            solutions = scores.solutions_a_priori(metric, solutions_a_priori, dominated)
            solutions.metadata["prior-solution-from"] = str(solution_fn)

            output_filename = (
                output_path / f"{score_fn.stem}@{solution_fn.stem}"
            ).with_suffix(".json")

        if deduplicate:
            logger.info(f"Deduplicating solutions from {output_filename}")
            solutions = solutions.deduplicate()

        utils.save_json_with_backup(output_filename, solutions.model_dump())
        logger.info(f"Saved `{output_filename}`")



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later


from fairical import utils


def test_make_table_basic() -> None:
    indicators: dict[str, dict[utils.IndicatorType, float]] = {
        "System A": {
            "hv": 0.8,
            "ud": 0.2,
            "os": 0.3,
            "as": 0.75,
            "onvg": 12,
            "onvgr": 0.6,
            "relative-onvg": 5,
            "area": 0.55,
        },
        "System B": {
            "hv": 0.6,
            "ud": 0.1,
            "os": 0.5,
            "as": 0.95,
            "onvg": 11,
            "onvgr": 0.5,
            "relative-onvg": 4,
            "area": 0.45,
        },
    }
    table = utils.make_table(indicators, fmt="simple")
    assert "System" in table
    assert "System A" in table
    assert "System B" in table
    assert "HV" in table
    assert "Area" in table
    # Check numeric formatting
    assert ".80" in table or "0.80" in table  # may vary by tabulate version


def test_make_table_dict_input() -> None:
    indicators: dict[str, dict[utils.IndicatorType, float]] = {
        "Sys1": {
            "hv": 0.7,
            "ud": 0.25,
            "os": 0.4,
            "as": 0.72,
            "onvg": 10,
            "onvgr": 0.55,
            "relative-onvg": 4,
            "area": 0.35,
        }
    }
    table = utils.make_table(indicators, fmt="github")
    assert "Sys1" in table
    assert (
        "|   Area |" in table or "| Area |" in table
    )  # allow for tabulate header formatting


def test_make_table_with_custom_keys() -> None:
    indicators: dict[str, dict[utils.IndicatorType, float]] = {
        "SysX": {
            "hv": 0.8,
            "ud": 0.2,
            "os": 0.3,
            "as": 0.9,
            "onvg": 12,
            "onvgr": 0.6,
            "relative-onvg": 5,
        },
        "SysY": {
            "hv": 0.6,
            "ud": 0.1,
            "os": 0.5,
            "as": 0.85,
            "onvg": 11,
            "onvgr": 0.5,
            "relative-onvg": 4,
        },
    }
    # Only use a subset of keys
    keys = ["hv", "ud", "os"]
    table = utils.make_table(indicators, fmt="simple", table_keys=keys)

    assert "HV" in table
    assert "UD" in table
    assert "OS" in table
    assert "Area" in table

    assert "ONVG" not in table
    assert "ONVGR" not in table
    assert "RELATIVE-ONVG" not in table



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Evaluation CLI for demographically fair machine learning systems.

This module defines the ``evaluate`` command, which loads one or more system solution
files in JSON format, validates them against a predefined data model, and computes
Pareto-front indicators. It supports tabulated output in various formats (via
`tabulate`) and optional generation of radar plots to visualize trade-offs between
metrics.

The CLI provides options to output formatting, and radar plot generation, and is
intended to support analysis of fairness-performance trade-offs across systems.
"""

import logging
import pathlib

logger = logging.getLogger(__name__.split(".", 1)[0])

import click
import matplotlib

# avoids loading X11-dependent backends
matplotlib.use("agg")

from ..solutions import Solutions
from . import utils


@click.command(
    epilog="""Examples:

\b
  1. Evaluate adjustable fairness (estimate Pareto/Non-dominated-solutions) from
     two systems, and print table with indicators:

     .. code:: sh

        fairical evaluate system1.json system2.json

\b
  2. Evaluate adjustable fairness and create a file with a radar plot (called
     ``radar.pdf``):

     .. code:: sh

        fairical evaluate system1.json system2.json --radar radar.pdf

""",
)
@click.argument(
    "system",
    type=click.Path(exists=True, dir_okay=False, readable=True, path_type=pathlib.Path),
    required=True,
    nargs=-1,
    callback=utils.validate_solutions_json,
)
@click.option(
    "-r",
    "--radar",
    type=click.Path(dir_okay=False, writable=True, path_type=pathlib.Path),
    help="Creates radar plot of system indicators and saves it to file",
)
@click.option(
    "-d/-D",
    "--deduplicate/--no-deduplicate",
    default=True,
    help="If we should prune (deduplicate) similar solutions.",
)
@click.option(
    "-t",
    "--table-format",
    type=click.Choice(__import__("tabulate").tabulate_formats),
    help="Output table format.",
    default="simple",
    show_default=True,
)
@utils.verbosity_option(logger)
def evaluate(
    system: dict[str, Solutions],
    radar: pathlib.Path | None,
    deduplicate: bool,
    table_format: str,
    verbose: int,
) -> None:  # numpydoc ignore=PR01
    """Evaluate systems for adjustable fairness.

    This command takes an arbitrary number of JSON files describing ML system solutions
    and estimates the Pareto front (i.e. non-dominated solutions according to an
    arbitrary number of utility and demographic fairness metrics), and outputs
    indicators that characterize the trade-off between all metrics.

    The JSON files must be formatted as described in the documentation.
    """
    # 1. for each system, get indicators
    indicator_values = []
    for k, v in system.items():
        logger.info(f"Deduplicating solutions from {k}")
        sol = v.deduplicate() if deduplicate else v
        indicator_values.append(sol.indicators())
    indicators = dict(zip(system.keys(), indicator_values))

    # 2. table and output to stdout
    from fairical.utils import make_table

    click.echo(make_table(indicators, fmt=table_format))

    if radar is not None:
        # [optional] 3. save radar plot of indicators
        from fairical.plot import radar_chart

        fig, _ = radar_chart(indicators)
        fig.savefig(radar)



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later


import json
import pathlib

import pydantic
import pytest

from fairical.scores import _CACHE, Scores
from fairical.solutions import Solutions


@pytest.fixture(autouse=True)
def clear_cache():
    """Clear the JSON cache before and after each test."""
    _CACHE.clear()
    yield
    _CACHE.clear()


def test_inline_valid():
    payload = {
        "scores": [[0.0, 1.0, 0.5], [0.2, 0.8, 0.3]],
        "identifiers": ["model-1", "model-2"],
        "ground-truth": [0, 1, 0],
        "attributes": {
            "race": ["A", "B", "C"],
            "gender": ["F", "M", "F"],
            "age": [20, 30, 40],
        },
    }

    dm = Scores.model_validate(payload)
    assert len(dm.scores) == 2
    assert isinstance(dm.scores, list)
    assert all(isinstance(s, list) for s in dm.scores)
    assert len(dm.identifiers) == len(dm.scores)
    assert dm.identifiers == ["model-1", "model-2"]
    assert dm.scores[0] == [0.0, 1.0, 0.5]
    assert isinstance(dm.ground_truth, list)
    assert dm.ground_truth == [0, 1, 0]
    assert isinstance(dm.attributes, dict)
    assert dm.attributes["age"] == [20, 30, 40]


def test_file_backed_valid(tmp_path: pathlib.Path):
    # Prepare JSON files
    scores_data = [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]
    gt_data = [0, 1, 0]
    attrs_data = {
        "race": ["X", "Y", "Z"],
        "gender": ["M", "F", "M"],
        "age": [25, 35, 45],
    }

    scores_file0 = tmp_path / "scores0.json"
    scores_file1 = tmp_path / "scores1.json"
    gt_file = tmp_path / "gt.json"
    attrs_file = tmp_path / "attrs.json"

    # dumps data
    scores_file0.write_text(json.dumps(scores_data[0]))
    scores_file1.write_text(json.dumps(scores_data[1]))
    gt_file.write_text(json.dumps(gt_data))
    attrs_file.write_text(json.dumps(attrs_data))
    payload = {
        "scores": [scores_file0, scores_file1],
        "ground-truth": gt_file,
        "attributes": attrs_file,
    }
    dm = Scores.model_validate(payload)

    # Cache should have entries for each path
    assert scores_file0.resolve() in _CACHE
    assert scores_file1.resolve() in _CACHE
    assert gt_file.resolve() in _CACHE
    assert attrs_file.resolve() in _CACHE

    # Scores fields loaded correctly
    assert all(isinstance(s, list) for s in dm.scores)
    assert dm.scores == scores_data
    assert dm.ground_truth == gt_data
    assert dm.attributes == attrs_data

    # Now test relative paths with context
    payload = {
        "scores": [scores_file0.name, scores_file1.name],
        "ground-truth": gt_file.name,
        "attributes": attrs_file.name,
    }
    main_file = tmp_path / "main.json"
    with main_file.open("w") as f:
        f.write(json.dumps(payload))
        f.flush()
    # Validate using model_validate_json and context for base_dir
    dm2 = Scores.load(main_file)

    # The nested files should be loaded correctly via relative paths
    assert all(isinstance(s, list) for s in dm2.scores)
    assert dm2.scores == scores_data
    assert dm2.ground_truth == gt_data
    assert dm2.attributes == attrs_data


def test_score_value_error():
    payload = {
        "scores": [[-0.1, 0.5, 1.1]],  # out-of-range
        "ground-truth": [0, 1, 0],
        "attributes": {
            "race": ["A", "B", "C"],
            "gender": ["F", "M", "F"],
            "age": [20, 30, 40],
        },
    }
    with pytest.raises(pydantic.ValidationError):
        Scores.model_validate(payload)


def test_ground_truth_value_error():
    payload = {
        "scores": [[0.2, 0.4, 0.6]],
        "ground-truth": [0, -1, 1],  # negative label
        "attributes": {
            "race": ["A", "B", "C"],
            "gender": ["F", "M", "F"],
            "age": [20, 30, 40],
        },
    }
    with pytest.raises(pydantic.ValidationError):
        Scores.model_validate(payload)


def test_attribute_type_error():
    payload = {
        "scores": [[0.2, 0.4, 0.6]],
        "ground-truth": [0, 1, 0],
        "attributes": {
            "race": ["A", "B", "C"],
            "gender": ["F", "M", "F"],
            "age": [["twenty"], "thirty", "forty"],
        },
    }
    with pytest.raises(pydantic.ValidationError):
        Scores.model_validate(payload)


def test_inconsistent_scores_length():
    payload = {
        "scores": [
            [0.1, 0.2, 0.3],
            [0.4, 0.5],  # inconsistent length
        ],
        "ground-truth": [0, 1, 0],
        "attributes": {
            "race": ["A", "B", "C"],
            "gender": ["F", "M", "F"],
            "age": [20, 30, 40],
        },
    }
    with pytest.raises(ValueError) as exc:
        Scores.model_validate(payload)
    assert "scores[1] length 2 != 3" in str(exc.value)


def test_inconsistent_identifiers_length():
    payload = {
        "scores": [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]],
        "identifiers": ["model-1", "model-2"],  # too short
        "ground-truth": [0, 1, 0],
        "attributes": {
            "race": ["A", "B", "C"],
            "gender": ["F", "M", "F"],
            "age": [20, 30, 40],
        },
    }
    with pytest.raises(ValueError) as exc:
        Scores.model_validate(payload)
    assert "identifiers length 2 != 3" in str(exc.value)


def test_inconsistent_ground_truth_length():
    payload = {
        "scores": [[0.5, 0.6, 0.7]],
        "ground-truth": [0, 1],  # too short
        "attributes": {
            "race": ["A", "B", "C"],
            "gender": ["F", "M", "F"],
            "age": [20, 30, 40],
        },
    }
    with pytest.raises(ValueError) as exc:
        Scores.model_validate(payload)
    assert "ground-truth length 2 != 3" in str(exc.value)


def test_inconsistent_attribute_length():
    payload = {
        "scores": [[0.1, 0.2, 0.3]],
        "ground-truth": [0, 1, 0],
        "attributes": {
            "race": ["A", "B"],
            "gender": ["F", "M", "F"],
            "age": [20, 30, 40],
        },
    }
    with pytest.raises(ValueError) as exc:
        Scores.model_validate(payload)
    assert "attribute 'race' length 2 != 3" in str(exc.value)


def test_generated_identifiers():
    payload = {
        "scores": [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]],
        "ground-truth": [0, 1, 0],
        "attributes": {
            "race": ["A", "B", "C"],
            "gender": ["F", "M", "F"],
            "age": [20, 30, 40],
        },
    }
    dm = Scores.model_validate(payload)
    assert dm.identifiers == ["model-0", "model-1"]


def test_getitem_and_empty_lists():
    payload = {
        "scores": [[]],
        "ground-truth": [],
        "attributes": {"race": [], "gender": [], "age": []},
    }
    dm = Scores.model_validate(payload)
    assert isinstance(dm.scores[0], list)

    # Empty lists should work
    assert len(dm.scores[0]) == 0
    # __getitem__ should raise IndexError on empty
    with pytest.raises(IndexError):
        _ = dm.scores[0][0]

    # Test getitem on ground_truth and attributes
    assert isinstance(dm.ground_truth, list)
    with pytest.raises(IndexError):
        _ = dm.ground_truth[0]

    assert isinstance(dm.attributes, dict)
    with pytest.raises(KeyError):
        _ = dm.attributes["nonexistent"]


def test_valid_fromarray_complete():
    points = [[0.1, 0.9, 0.8], [0.3, 0.3, 0.7]]
    metrics = ("acc", "f1", "eod+race")
    identifiers = [0, 1]
    identifier_names = ["model-1", "model-2"]
    thresholds = [0.1, 0.2, 0.3]

    solutions = Solutions.fromarray(
        points=points,
        metrics=metrics,
        identifiers=identifiers,
        identifier_names=identifier_names,
        thresholds=thresholds,
    )

    assert solutions.points == {
        "acc": [0.1, 0.3],
        "f1": [0.9, 0.3],
        "eod+race": [0.8, 0.7],
    }
    assert solutions.metadata["identifier-names"] == identifier_names
    assert solutions.metadata["identifiers"] == identifiers
    assert solutions.metadata["thresholds"] == thresholds


def test_valid_fromarray_no_identifiers():
    points = [[0.1, 0.9, 0.8], [0.3, 0.3, 0.7]]
    metrics = ("acc", "f1", "eod+race")
    thresholds = [0.1, 0.2, 0.3]

    solutions = Solutions.fromarray(
        points=points,
        metrics=metrics,
        thresholds=thresholds,
    )

    assert solutions.points == {
        "acc": [0.1, 0.3],
        "f1": [0.9, 0.3],
        "eod+race": [0.8, 0.7],
    }
    assert list(solutions.metadata["identifiers"]) == [0, 0]
    assert solutions.metadata["identifier-names"] == ["model-0"]
    assert solutions.metadata["thresholds"] == thresholds


def test_valid_fromarray_no_identifier_names():
    points = [[0.1, 0.9, 0.8], [0.3, 0.3, 0.7]]
    metrics = ("acc", "f1", "eod+race")
    identifiers = [0, 1]
    thresholds = [0.1, 0.2, 0.3]

    solutions = Solutions.fromarray(
        points=points,
        metrics=metrics,
        identifiers=identifiers,
        thresholds=thresholds,
    )

    assert solutions.points == {
        "acc": [0.1, 0.3],
        "f1": [0.9, 0.3],
        "eod+race": [0.8, 0.7],
    }
    assert list(solutions.metadata["identifiers"]) == identifiers
    assert list(solutions.metadata["identifier-names"]) == [
        "model-0",
        "model-1",
    ]
    assert solutions.metadata["thresholds"] == thresholds


def test_invalid_solution_single():
    with pytest.raises(ValueError) as exc:
        Solutions.fromarray([[0.1, 0.9], [0.3, 0.3]], ("tpr", "eod+age"), [0.2, 0.25])
    assert "cannot use `tpr` as the sole utility metrics" in str(exc.value)

    with pytest.raises(ValueError) as exc:
        Solutions.fromarray([[0.1, 0.9], [0.3, 0.3]], ("tnr", "eod+age"), [0.2, 0.25])
    assert "cannot use `tnr` as the sole utility metrics" in str(exc.value)

    with pytest.raises(ValueError) as exc:
        Solutions.fromarray([[0.1, 0.9], [0.3, 0.3]], ("fpr", "eod+age"), [0.2, 0.25])
    assert "cannot use `fpr` as the sole utility metrics" in str(exc.value)

    with pytest.raises(ValueError) as exc:
        Solutions.fromarray([[0.1, 0.9], [0.3, 0.3]], ("fnr", "eod+age"), [0.2, 0.25])
    assert "cannot use `fnr` as the sole utility metrics" in str(exc.value)


def test_invalid_solution_pair():
    with pytest.raises(ValueError) as exc:
        Solutions.fromarray(
            [[0.1, 0.9, 0.8], [0.3, 0.3, 0.7]],
            ("fnr", "eod+age", "tpr"),
            [0.2, 0.25, 0.75],
        )
    assert "cannot use the pair `['fnr', 'tpr']` as the sole utility metrics" in str(
        exc.value
    )

    with pytest.raises(ValueError) as exc:
        Solutions.fromarray(
            [[0.1, 0.9, 0.8], [0.3, 0.3, 0.7]],
            ("tnr", "eod+age", "fpr"),
            [0.2, 0.25, 0.75],
        )
    assert "cannot use the pair `['tnr', 'fpr']` as the sole utility metrics" in str(
        exc.value
    )

    with pytest.raises(ValueError) as exc:
        Solutions.fromarray(
            [[0.1, 0.9, 0.8], [0.3, 0.3, 0.7]],
            ("rec", "eod+age", "fnr"),
            [0.2, 0.25, 0.75],
        )
    assert "cannot use the pair `['rec', 'fnr']` as the sole utility metrics" in str(
        exc.value
    )



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Visualization CLI for demographically fair machine learning systems.

This module defines the ``pareto`` command, which loads one or more system solutions in
JSON format and computes the pareto fronts, with options to generate visualization plots.

The CLI provides options to output formatting, and plot generation, and is
intended to support analysis of fairness-performance trade-offs across systems.
"""

import logging
import pathlib

logger = logging.getLogger(__name__.split(".", 1)[0])

import click
import matplotlib

# avoids loading X11-dependent backends
matplotlib.use("agg")

from ..solutions import Solutions
from . import utils


@click.command(
    epilog="""Examples:

\b
  1. Evaluate adjustable fairness and display the pareto front in table format on the terminal

     .. code:: sh

        fairical pareto system-1.json system-2.json

    2. Evaluate adjustable fairness display the pareto front in table format on the terminal,
       and generate a plot showing the dominated and non-dominated solutions.

     .. code:: sh

        fairical pareto system-1.json system-2.json --plot
""",
)
@click.argument(
    "system",
    type=click.Path(exists=True, dir_okay=False, readable=True, path_type=pathlib.Path),
    required=True,
    nargs=-1,
    callback=utils.validate_solutions_json,
)
@click.option(
    "-d/-D",
    "--deduplicate/--no-deduplicate",
    default=True,
    help="If we should prune (deduplicate) similar solutions.",
)
@click.option(
    "-p",
    "--plot",
    is_flag=True,
    default=False,
    help="If we should plot the solutions",
)
@click.option(
    "--hide-ds",
    is_flag=True,
    default=False,
    help="If true, hide the ds points for a-priori data from the plot.",
)
@click.option(
    "-o",
    "--output-path",
    type=click.Path(
        dir_okay=True, file_okay=False, writable=True, path_type=pathlib.Path
    ),
    help="Directory where to store system solutions.",
)
@utils.verbosity_option(logger)
def pareto(
    system: dict[str, Solutions],
    deduplicate: bool,
    plot: bool,
    hide_ds: bool,
    output_path: pathlib.Path | None,
    verbose: int,
) -> None:  # numpydoc ignore=PR01
    """Compute pareto frontiers of systems.

    This command takes an arbitrary number of JSON files describing ML system solutions
    and displays the Pareto front (i.e. non-dominated solutions according to an
    arbitrary number of utility and demographic fairness metrics) on the terminal.

    A scatter plot showing the dominated and non-dominated solutions can be created
    by passing the `--plot` flag.

    The JSON files must be formatted as described in the documentation.
    """
    from fairical.plot import pareto_plot

    if output_path is None:
        output_path = pathlib.Path()

    if deduplicate:
        for k, v in system.items():
            logger.info(f"Deduplicating solutions from {k}")
            system[k] = v.deduplicate()

    solutions = {k: v.non_dominated_solutions() for k, v in system.items()}
    if plot:
        fig, _ = pareto_plot(solutions, hide_ds=hide_ds)

        output_file = output_path / "pareto.pdf"
        logger.info(f"Saving plot at `{output_file}`...")
        utils.prepare_and_backup(output_file)
        fig.savefig(output_file)

    for label, (nds, _) in solutions.items():
        table = nds.tabulate()

        print(f"\nPareto front for {label}")
        print(table)



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy
import numpy.typing
import pytest

from fairical.solutions import Solutions


@pytest.mark.parametrize(
    "solutions, metrics, thresholds, manual, tol",
    [
        (
            # 2d case with mininimisation and maximisation
            # note: [0.4, 0.6] is dominated by the other 2 points
            [[0.2, 0.8], [0.4, 0.6], [0.3, 0.9]],
            ("eod+race", "acc"),
            [0.5, 0.5, 0.7],
            0.01,
            0.001,
        ),
        (
            # 3d case with mininimisation and maximisation
            [[0.3, 0.5, 0.9], [0.1, 0.9, 0.6], [0.2, 0.1, 0.3]],
            ("eod+race", "eod+age", "acc"),
            [0.2, 0.5, 0.4],
            0.096,
            0.001,
        ),
        (
            # 2d case with no spread
            [[0.5, 0.5], [0.5, 0.5]],
            ("eod+race", "acc"),
            [0.5, 0.5],
            0.0,
            0.001,
        ),
        (
            # 2d case with maximum spread
            [[0.0, 0.0], [1.0, 1.0]],
            ("eod+race", "acc"),
            [0.5, 0.5],
            1.0,
            0.001,
        ),
    ],
)
def test_os(
    solutions: numpy.typing.ArrayLike,
    metrics: tuple[str],
    thresholds: list[float],
    manual: float,
    tol: float,
) -> None:
    """A valid system with three points."""

    sol = Solutions.fromarray(solutions, metrics, thresholds)
    ind = sol.indicators()
    assert numpy.isclose(ind["os"], manual, tol)


@pytest.mark.parametrize(
    "solutions, metrics, thresholds, manual, tol",
    [
        (
            # 2d case with mininimisation and maximisation
            # note: [0.4, 0.6] is dominated by the other 2 points
            [[0.2, 0.8], [0.4, 0.6], [0.3, 0.9]],
            ("eod+race", "acc"),
            [0.5, 0.5, 0.7],
            0.1,
            0.001,
        ),
        (
            # 3d case with mininimisation and maximisation
            [[0.3, 0.5, 0.9], [0.1, 0.9, 0.6], [0.2, 0.1, 0.3]],
            ("eod+race", "eod+age", "acc"),
            [0.2, 0.5, 0.4],
            0.533333,
            0.001,
        ),
        (
            # 2d case with no spread
            [[0.5, 0.5], [0.5, 0.5]],
            ("eod+race", "acc"),
            [0.5, 0.5],
            0.0,
            0.001,
        ),
        (
            # 2d case with maximum spread
            [[0.0, 0.0], [1.0, 1.0]],
            ("eod+race", "acc"),
            [0.5, 0.5],
            1.0,
            0.001,
        ),
    ],
)
def test_as(
    solutions: numpy.typing.ArrayLike,
    metrics: tuple[str],
    thresholds: list[float],
    manual: float,
    tol: float,
) -> None:
    """A valid system with three points."""

    sol = Solutions.fromarray(solutions, metrics, thresholds)
    ind = sol.indicators()
    assert numpy.isclose(ind["as"], manual, tol)



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Data model organizing scores of ML systems under multi-objective constraints.

The model describes a data structure containing machine learning model scores,
ground-truth labels, and sensitive attributes (e.g., race, gender).
"""

import itertools
import json
import logging
import operator
import pathlib
import typing

import pydantic
import sklearn.metrics

from . import metrics, solutions

logger = logging.getLogger(__name__)

_CACHE: dict[pathlib.Path, list] = {}
"""We read input files only once, and then serve repeated reads from the cache to save
memory and read-time."""


def _cache_json(path: pathlib.Path) -> list | dict:
    """Load and cache a JSON file.

    Parameters
    ----------
    path
        Path to the file to load and cache.

    Returns
    -------
    list
        A list of scores, ground-truth or attributes.
    """
    path = path.resolve()
    if path in _CACHE:
        return _CACHE[path]
    return _CACHE.setdefault(path, json.load(path.open("r")))


def _single_system_thresholds(
    scores: typing.Sequence[float],
    ground_truth: typing.Sequence[int],
    eval_metrics: typing.Sequence[str],
) -> list[float]:
    """Calculate a set of thresholds to use for single system evaluation.

    Helper method for py:func:`solutions` -- check its documentation for help.

    Parameters
    ----------
    scores
        Scores of the sub-system being analyzed.
    ground_truth
        Ground-truth for the task.
    attributes
        Protected attributes to consider for the calculation of demographic fairness
        metrics.
    thresholds
        List of thresholds to apply as values within the interval :math:`[0,1]`. If not
        provided, then estimate thresholds using scikit's ROC technique.

    Returns
    -------
        All know solutions for the input system.
    """

    if any([k in eval_metrics for k in ("prec", "rec", "avg_prec", "f1")]):
        # use precision-recall-thresholds as thresholds
        _, _, pr_thresholds = sklearn.metrics.precision_recall_curve(
            ground_truth, scores
        )
        return list(pr_thresholds)

    # else any([k in eval_metrics for k in ("fpr", "tpr", "roc_auc", "acc")]):
    # use roc-thresholds as thresholds
    _, _, roc_thresholds = sklearn.metrics.roc_curve(ground_truth, scores)
    return list(roc_thresholds)


def _single_system_solutions(
    scores: typing.Sequence[float],
    ground_truth: typing.Sequence[int],
    attributes: typing.Mapping[str, list[int | str]],
    eval_metrics: typing.Sequence[str],
    thresholds: list[float],
) -> dict[str, list[float]]:
    """Calculate all solutions of a sub-system, given metrics and thresholds.

    Helper method for py:func:`solutions` -- check its documentation for help.

    Parameters
    ----------
    scores
        Scores of the sub-system being analyzed.
    ground_truth
        Ground-truth for the task.
    attributes
        Protected attributes to consider for the calculation of demographic fairness
        metrics.
    eval_metrics
        Metrics types to consider when evaluating solutions. Example: ``eod+age``,
        ``eod+gender``, or ``acc``.
    thresholds
        List of thresholds to apply as values within the interval :math:`[0,1]`. If not
        provided, then estimate thresholds using scikit's ROC technique.

    Returns
    -------
        All know solutions for the input system.
    """

    retval: dict[str, list[float]] = {}

    for metric in eval_metrics:
        retval[metric] = metrics.calculate_metric(
            metric, ground_truth, scores, thresholds, attributes
        )

    return retval


class Scores(pydantic.BaseModel):
    """Data model representing raw machine learning score outputs.

    It is composed of a set of scores, for one or more operating points (e.g. preference
    rays, or ratios between various optimisation objectives), ground-truth for the task
    being analyzed, as well as extra (protected?) attributes that are relevant for, at
    least, demographic fairness analysis.

    For the JSON representation, scores, ground-truth, and demographic attributes may be
    inlined or out-sourced to an external file where the data structure can be loaded
    from.  Relative paths are considered w.r.t. the location of the current file.
    """

    #: Inline scores data or list of file paths. Each score must be a floating-point
    #: number between 0 and 1 inclusive.
    scores: list[
        list[typing.Annotated[float, pydantic.Field(ge=0.0, le=1.0)]] | pathlib.Path
    ] = pydantic.Field(
        ...,
        description="One or many lists of scores representing different operating "
        "points of the system being analyzed.",
    )

    #: Optional inline identifiers corresponding to each system in scores. Each identifier must be a string.
    identifiers: list[str] | None = pydantic.Field(
        default=None,
        description="Optional list of names for each item in scores.",
    )

    #: Inline ground-truth data or a single file path. Each ground-truth label must be
    #: an integer with a minimum value of 0.
    ground_truth: list[typing.Annotated[int, pydantic.Field(ge=0)]] | pathlib.Path = (
        pydantic.Field(
            ...,
            alias="ground-truth",
            description="Ground-truth for the task being analyzed.",
        )
    )

    #: Inline attributes data or a single file path. It is setup as a dictionary
    #: mapping attribute names to lists of demographic data, which can be of type str,
    #: integer or floating-point.
    attributes: (
        dict[
            str,
            list[str | typing.Annotated[int, pydantic.Field(ge=0)]],
        ]
        | pathlib.Path
    ) = pydantic.Field(
        ...,
        description="Meta-data (demographic or other) attributes for samples "
        "in the task.",
    )

    @pydantic.model_validator(mode="after")
    def maybe_load_members(self, info: pydantic.ValidationInfo) -> typing.Self:
        """Load all external files if needed."""

        base = pathlib.Path.cwd()
        if info.context is not None:
            base = pathlib.Path(info.context.get("base_dir", base))

        for k, v in enumerate(self.scores):
            if isinstance(v, pathlib.Path):
                path = (base / v) if not v.is_absolute() else v
                self.scores[k] = typing.cast(list[float], _cache_json(path))

        if isinstance(self.ground_truth, pathlib.Path):
            gt_path = (
                (base / self.ground_truth)
                if not self.ground_truth.is_absolute()
                else self.ground_truth
            )
            self.ground_truth = typing.cast(list[int], _cache_json(gt_path))

        if isinstance(self.attributes, pathlib.Path):
            attr_path = (
                (base / self.attributes)
                if not self.attributes.is_absolute()
                else self.attributes
            )
            self.attributes = typing.cast(
                dict[str, list[int | str]], _cache_json(attr_path)
            )

        return self

    @pydantic.model_validator(mode="after")
    def check_consistent_num_samples(self) -> typing.Self:
        """Ensure all sample-level lists have the same length."""

        assert isinstance(self.scores[0], list)
        expected = len(self.scores[0])

        # Validate each ScoreList
        for idx, score_list in enumerate(self.scores):
            assert isinstance(score_list, list)
            if len(score_list) != expected:
                raise ValueError(
                    f"scores[{idx}] length {len(score_list)} != {expected}"
                )

        # Validate ground-truth length
        assert isinstance(self.ground_truth, list)
        gt_len = len(self.ground_truth)
        if gt_len != expected:
            raise ValueError(f"ground-truth length {gt_len} != {expected}")

        # Validate attributes lengths
        assert isinstance(self.attributes, dict)
        for name, values in self.attributes.items():
            if len(values) != expected:
                raise ValueError(
                    f"attribute '{name}' length {len(values)} != {expected}"
                )

        return self

    @pydantic.model_validator(mode="after")
    def check_identifiers_specified(self) -> typing.Self:
        """Generate default identifier per system if not specified."""
        if self.identifiers is None:
            logger.info(
                "No identifiers specified, they will be automatically generated."
            )

            self.identifiers = [f"model-{n}" for n in range(len(self.scores))]
        return self

    @pydantic.model_validator(mode="after")
    def check_consistent_num_subsystems(self) -> typing.Self:
        """Ensure all subsystem-level lists have the same length."""

        assert isinstance(self.scores, list)
        expected = len(self.scores)

        # Validate identifiers length
        assert isinstance(self.identifiers, list)
        identifiers_len = len(self.identifiers)
        if identifiers_len != expected:
            raise ValueError(f"identifiers length {identifiers_len} != {expected}")

        return self

    @classmethod
    def load(cls, source: pathlib.Path | str | typing.TextIO) -> typing.Self:
        """Validate and load a JSON file into a raw data object.

        This function is intended to validate and load the input in JSON format. It opens
        the given file path, parses its JSON content, and validates it against the defined
        data model.

        Parameters
        ----------
        source
            Source input where to read JSON from.

        Returns
        -------
            Parsed and validated content as a :py:class:`Scores` instance.

        Raises
        ------
        pydantic_core.ValidationError
            If the file contains invalid data.
        """

        if isinstance(source, pathlib.Path | str):
            path = pathlib.Path(source)
            return cls.model_validate_json(
                path.read_text(), context={"base_dir": path.parent}
            )

        else:  # noqa: RET505
            return cls.model_validate_json(source.read())

    def save(self, dest: pathlib.Path | str | typing.TextIO, **args) -> None:
        """Save contents to an external file.

        Parameters
        ----------
        dest
            Destination where to save the contents. If not a path or str, then assumed
            to have a ``write`` method accepting strings.
        args
            Parameters further passed down to
            :py:func:`pydantic.BaseModel.model_dump_json`.
        """

        if isinstance(dest, pathlib.Path | str):
            with pathlib.Path(dest).open("w", encoding="utf-8") as f:
                f.write(self.model_dump_json(**args))

        else:
            dest.write(self.model_dump_json(**args))

    def solutions_a_posteriori(
        self,
        metrics: typing.Sequence[str],
        thresholds: list[float] | None = None,
    ) -> solutions.Solutions:
        """Calculate all solutions of a system *a posteriori*, given metrics and thresholds.

        This method retrieves solutions that can be implemented by systems. For each set
        of scores in ``self.scores``, it calculates all solutions of the system being
        analysed through simple thresholding, and then aggregates all solutions to
        construct all possible sets of solutions a system can implement.

        Parameters
        ----------
        metrics
            Metrics types to consider when evaluating solutions. Example: ``eod+age``,
            ``eod+gender``, or ``acc``.
        thresholds
            List of thresholds to apply as values within the interval :math:`[0,1]`. If
            not provided, then uses scikit-learn to extract meaningful scores from the
            system.

        Returns
        -------
            All know solutions for the input system.
        """

        points: dict[str, list[float]] = {k: [] for k in metrics}
        metadata: dict[str, typing.Any] = {
            "identifiers": [],
            "thresholds": [],
        }

        for i, subsystem in enumerate(self.scores):
            logger.info(f"Evaluating subsystem {i + 1}/{len(self.scores)}")

            thresholds = thresholds or _single_system_thresholds(
                typing.cast(typing.Sequence[float], subsystem),
                typing.cast(typing.Sequence[int], self.ground_truth),
                metrics,
            )

            logger.info(
                f"Using {len(thresholds)} thresholds for evaluation of solutions."
            )

            # a-posteriori thresholds
            sol = _single_system_solutions(
                # note: at this point we are sure not to have any path lying around!
                typing.cast(typing.Sequence[float], subsystem),
                typing.cast(typing.Sequence[int], self.ground_truth),
                typing.cast(typing.Mapping[str, list[int | str]], self.attributes),
                metrics,
                thresholds,
            )
            metadata["thresholds"] += thresholds
            metadata["identifier-names"] = self.identifiers
            for k in metrics:
                points[k] += sol[k]

            metadata["identifiers"] += itertools.repeat(i, len(list(sol.values())[0]))

        return solutions.Solutions.model_validate(
            {"points": points, "metadata": metadata}
        )

    def solutions_a_priori(
        self,
        metrics: typing.Sequence[str],
        prior_solutions: solutions.Solutions,
        dominated: bool | None = None,
    ) -> solutions.Solutions:
        """Calculate all solutions of a system with settings (system and metrics) *a priori*.

        This method retrieves solutions that can be implemented by systems. For each set of
        scores in ``self.scores``, it calculates all solutions of the system being
        analysed through simple thresholding, and then aggregates all solutions to
        construct all possible sets of solutions a system can implement.

        Parameters
        ----------
        metrics
            Metrics types to consider when evaluating solutions. Example: ``eod+age``,
            ``eod+gender``, or ``acc``.
        prior_solutions
            Solutions to use as the basis for the calculation of solutions with
            threshold and system picked *a priori*.
        dominated
            A tri-state boolean flag that defines which subset of ``prior_solutions`` to
            apply to the scores:

            * ``None``: applies all solutions (default)
            * ``True``: applies only dominated solutions
            * ``False``: applies only non-dominated solutions

        Returns
        -------
            All *a priori* known solutions for the input system.
        """

        points: dict[str, list[float]] = {k: [] for k in metrics}

        metadata = prior_solutions.metadata
        metadata["prior-solution-set"] = "all"

        if dominated in (True, False):
            nds, ds = prior_solutions.non_dominated_solutions()
            if dominated:
                logger.info("Applying only dominated solutions...")
                prior_solutions = ds
                metadata = ds.metadata
                metadata["prior-solution-set"] = "dominated"
            else:
                logger.info("Applying only non-dominated solutions...")
                prior_solutions = nds
                metadata = nds.metadata
                metadata["prior-solution-set"] = "non-dominated"

        logger.info(f"Evaluating {len(metadata['thresholds'])} operating modes...")

        # for each set of **continguous** identifiers and associated thresholds, compute
        # metrics with those settings
        for identifier, group in itertools.groupby(
            zip(metadata["identifiers"], metadata["thresholds"]), operator.itemgetter(0)
        ):
            sol = _single_system_solutions(
                typing.cast(typing.Sequence[float], self.scores[identifier]),
                typing.cast(typing.Sequence[int], self.ground_truth),
                typing.cast(typing.Mapping[str, list[int | str]], self.attributes),
                metrics,
                list(map(operator.itemgetter(1), group)),
            )
            for k in metrics:
                points[k] += sol[k]

        return solutions.Solutions.model_validate(
            {"points": points, "metadata": metadata}
        )



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib
import re

from click.testing import CliRunner

from fairical.scripts.cli import cli


def str_counter(substr: str, s: str) -> int:
    """Count number of occurences of regular expression ``str`` on ``s``.

    Parameters
    ----------
    substr
        String or regular expression to search for in ``s``.
    s
        String where to search for ``substr``, that may include new-line characters.

    Returns
    -------
        The count on the number of occurences of ``substr`` on ``s``.
    """

    return sum(1 for _ in re.finditer(substr, s, re.MULTILINE))


def test_cli_help() -> None:
    runner = CliRunner()
    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0
    assert result.output.startswith("Usage:")


def test_cli_solve_help() -> None:
    runner = CliRunner()
    result = runner.invoke(cli, ["solve", "--help"])
    assert result.exit_code == 0
    assert result.output.startswith("Usage:")


def test_cli_evaluate_help() -> None:
    runner = CliRunner()
    result = runner.invoke(cli, ["evaluate", "--help"])
    assert result.exit_code == 0
    assert result.output.startswith("Usage:")


def test_cli_pareto_help() -> None:
    runner = CliRunner()
    result = runner.invoke(cli, ["pareto", "--help"])
    assert result.exit_code == 0
    assert result.output.startswith("Usage:")


def test_cli_solve(datadir: pathlib.Path, tmp_path: pathlib.Path, caplog):
    runner = CliRunner()

    with caplog.at_level("INFO", logger="fairical"):
        result = runner.invoke(
            cli,
            [
                "solve",
                "-vv",
                "--metric=acc",
                "--metric=eod+race",
                "--metric=eod+gender",
                str(datadir / "data" / "empirical" / "system_1.json"),
                str(datadir / "data" / "empirical" / "system_2.json"),
                "--output-path",
                str(tmp_path),
                "--thresholds=3",
            ],
        )

    assert result.exit_code == 0

    assert (
        str_counter("Calculating solutions \\*a posteriori\\* for system", caplog.text)
        == 2
    )
    assert str_counter("Evaluating subsystem", caplog.text) == 50
    assert str_counter("Deduplication reduced solutions", caplog.text) == 2
    assert str_counter("Saving only non-dominated solutions", caplog.text) == 0
    assert str_counter("Saving only dominated solutions", caplog.text) == 0
    assert str_counter(f"Saved `{str(tmp_path)}.*`$", caplog.text) == 2

    assert pathlib.Path(tmp_path / "system_1.json").exists()
    assert pathlib.Path(tmp_path / "system_2.json").exists()


def test_cli_solve_a_priori(datadir: pathlib.Path, tmp_path: pathlib.Path, caplog):
    runner = CliRunner()

    with caplog.at_level("INFO", logger="fairical"):
        result = runner.invoke(
            cli,
            [
                "solve",
                "-vv",
                "--metric=acc",
                "--metric=eod+race",
                "--metric=eod+gender",
                str(datadir / "data" / "empirical" / "system_1.json")
                + "@"
                + str(datadir / "data" / "empirical" / "system_2_nds.json"),
                "--output-path",
                str(tmp_path),
                "--thresholds=3",
            ],
        )

    assert result.exit_code == 0

    assert pathlib.Path(tmp_path / "system_1@system_2_nds.json").exists()

    assert (
        str_counter("Calculating solutions \\*a priori\\* for system", caplog.text) == 1
    )
    assert str_counter("Evaluating 9 operating modes", caplog.text) == 1
    assert str_counter("Deduplication reduced solutions", caplog.text) == 1
    assert str_counter("Saving only non-dominated solutions", caplog.text) == 0
    assert str_counter("Saving only dominated solutions", caplog.text) == 0
    assert str_counter(f"Saved `{str(tmp_path)}.*`$", caplog.text) == 1


def test_cli_evaluate(datadir: pathlib.Path, tmp_path: pathlib.Path, caplog):
    runner = CliRunner()

    radar_pdf = tmp_path / "radar.pdf"

    with caplog.at_level("INFO", logger="fairical"):
        result = runner.invoke(
            cli,
            [
                "evaluate",
                "-vv",
                str(datadir / "data" / "uc-2" / "system_1.json"),
                str(datadir / "data" / "uc-2" / "system_2.json"),
                "--deduplicate",
                "--radar",
                str(radar_pdf),
            ],
        )

    assert result.exit_code == 0
    assert radar_pdf.exists()

    assert (
        str_counter(
            r"^\s*system_1\s*0.12\s*0.50\s*1.00\s*0.00\s*0.24\s*0.12$", result.output
        )
        == 1
    )
    assert (
        str_counter(
            r"^\s*system_2\s*1.00\s*0.89\s*1.00\s*0.26\s*0.09\s*0.43$", result.output
        )
        == 1
    )

    assert str_counter("Loading solutions for system", caplog.text) == 2
    assert str_counter("Deduplication reduced solutions", caplog.text) == 2
    assert str_counter("Non-dominated/Dominated solutions", caplog.text) == 2


def test_cli_evaluate_nodedup(datadir: pathlib.Path, tmp_path: pathlib.Path, caplog):
    runner = CliRunner()

    radar_pdf = tmp_path / "radar.pdf"

    with caplog.at_level("INFO", logger="fairical"):
        result = runner.invoke(
            cli,
            [
                "evaluate",
                "-vv",
                str(datadir / "data" / "uc-2" / "system_1.json"),
                str(datadir / "data" / "uc-2" / "system_2.json"),
                "--no-deduplicate",
            ],
        )

    assert result.exit_code == 0
    assert not radar_pdf.exists()  # should not exist, as no opt given

    assert (
        str_counter(
            r"^\s*system_1\s*0.12\s*0.04\s*1.00\s*0.00\s*0.24\s*0.02$", result.output
        )
        == 1
    )
    assert (
        str_counter(
            r"^\s*system_2\s*1.00\s*0.32\s*1.00\s*0.26\s*0.09\s*0.20$", result.output
        )
        == 1
    )

    assert str_counter("Loading solutions for system.*", caplog.text) == 2
    assert str_counter("Non-dominated/Dominated solutions", caplog.text) == 2


def test_cli_pareto_3d(datadir: pathlib.Path, tmp_path: pathlib.Path, caplog):
    runner = CliRunner()

    with caplog.at_level("INFO", logger="fairical"):
        result = runner.invoke(
            cli,
            [
                "pareto",
                "-vv",
                str(datadir / "data" / "uc-3" / "system_1.json"),
                str(datadir / "data" / "uc-3" / "system_2.json"),
                "--plot",
                "--output-path",
                str(tmp_path),
            ],
        )

    expected_pdf_path = tmp_path / "pareto.pdf"
    assert result.exit_code == 0
    assert pathlib.Path(expected_pdf_path).exists()

    assert str_counter("Loading solutions for system", caplog.text) == 2
    assert str_counter("Deduplication reduced solutions", caplog.text) == 2
    assert str_counter("Non-dominated/Dominated solutions", caplog.text) == 2
    assert str_counter(f"Saving plot at `{expected_pdf_path}`...", caplog.text) == 1



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy
import pytest

from fairical.utils import normalized_radar_area


@pytest.mark.parametrize(
    "values,expected",
    [
        # trivial use-cases
        ([1, 1, 1], 1),
        ([1, 1, 1, 1], 1),
        ([1, 1, 1, 1, 1], 1),
        ([0, 0, 0], 0),
        ([0, 0, 0, 0], 0),
        ([0, 0, 0, 0, 0], 0),
        ([1, 0, 0], 0),
        ([1, 1, 0], 1 / 3),
        ([1, 1, 1, 0], 2 / 4),
        ([1, 1, 1, 0, 0], 2 / 5),
        ([1, 1, 1, 1, 0], 3 / 5),
        ([1, 1, 0, 1, 1], 3 / 5),
        ([0.5, 0.5, 0.5], 1 / 4),
        ([0.5, 0.5, 0.5, 0.5], 1 / 4),
        ([0.5, 0.5, 0.5, 0.5, 0.5], 1 / 4),
    ],
    ids=lambda x: str(x),  # just changes how pytest prints it
)
def test_case(values: list[float], expected: float):
    assert numpy.allclose(normalized_radar_area(values), expected)



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy
import pytest

from fairical.solutions import _hypervolume as hv


def _manual_hv_2d(points: numpy.ndarray, ref: numpy.ndarray) -> float:
    """Compute 2‑D hyper‑volume manually for small sets (helper)."""
    # brute‑force via inclusion–exclusion for ≤2 points
    if len(points) == 1:
        return numpy.prod(ref - points[0])
    if len(points) == 2:
        a1 = numpy.prod(ref - points[0])
        a2 = numpy.prod(ref - points[1])
        # overlap rectangle upper‑left corner is component‑wise max
        overlap_low = numpy.maximum(points[0], points[1])
        overlap = numpy.prod(numpy.maximum(ref - overlap_low, 0))
        return a1 + a2 - overlap
    raise ValueError("Helper only supports up to 2 points.")


@pytest.mark.parametrize(
    "pts, ref, manual",
    [
        (numpy.array([[1.0, 1.0]]), numpy.array([3.0, 3.0]), 4.0),
        (numpy.array([[1.0, 1.0], [1.0, 1.0]]), numpy.array([3.0, 3.0]), 4.0),
        (numpy.array([[1.0, 2.0], [2.0, 1.0]]), numpy.array([3.0, 3.0]), 3.0),
    ],
)
def test_hv_2d_exact(pts: numpy.ndarray, ref: numpy.ndarray, manual: float) -> None:
    """Exact hyper‑volume for 1‑ or 2‑point 2‑D sets matches manual calc."""
    expected = _manual_hv_2d(pts, ref)
    assert numpy.isclose(manual, expected)
    assert numpy.isclose(hv(pts, ref), expected)


def test_hv_3d_single_point() -> None:
    """Single point 3‑D volume is (ref‑point).prod()."""
    pts = numpy.array([[1.0, 1.0, 1.0]])
    ref = numpy.array([2.0, 2.0, 2.0])
    expected = numpy.prod(ref - pts[0])
    assert numpy.isclose(hv(pts, ref), expected)


def test_hv_dimension_mismatch() -> None:
    """Mismatch between nds objectives and ref‑point length raises error."""
    pts = numpy.array([[1.0, 2.0]])
    ref = numpy.array([3.0, 3.0, 3.0])  # one extra dimension
    with pytest.raises(ValueError):
        hv(pts, ref)


def test_hv_ref_point_not_1d() -> None:
    """Supplying a 2‑D ref‑point must raise, enforcing 1‑D requirement."""
    pts = numpy.array([[1.0, 2.0]])
    ref = numpy.array([[3.0, 3.0]])  # 2‑D instead of 1‑D
    with pytest.raises(ValueError):
        hv(pts, ref)



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Command-line interface for assessing demographically fair ML systems.

This module defines the root CLI entry point using :py:mod:`click`. It includes a base
command group and dynamically registers subcommands by importing them from other modules
in the package.
"""

import importlib

import click


@click.group(
    context_settings=dict(help_option_names=["-?", "-h", "--help"]),
)
def cli():
    """Assess adjustable demographically fair Machine Learning (ML) systems."""
    pass


def _add_command(module, obj):
    cli.add_command(
        getattr(importlib.import_module("." + module, package=__name__), obj)
    )


_add_command(".solve", "solve")
_add_command(".evaluate", "evaluate")
_add_command(".pareto", "pareto")



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import typing

import numpy

import fairical.metrics

# Example extracted from:
# https://fairlearn.org/v0.12/user_guide/assessment/common_fairness_metrics.html#common-fairness-metrics
y_true = [0, 1, 1, 1, 1, 0, 1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1]
y_pred = [0, 0, 1, 0, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0]
sf_data = [
    "b",
    "b",
    "a",
    "b",
    "b",
    "c",
    "c",
    "c",
    "a",
    "a",
    "c",
    "a",
    "b",
    "c",
    "c",
    "b",
    "c",
    "c",
]


def test_dpd():
    r = fairical.metrics.calculate_metric(
        "dpd+foo",
        y_true,
        y_pred,
        thresholds=[0.5],
        sensitive_attributes={"foo": sf_data},
    )

    assert len(r) == 1
    assert numpy.isclose(r, 0.25)


def test_dpr():
    r = fairical.metrics.calculate_metric(
        "dpr+foo",
        y_true,
        y_pred,
        thresholds=[0.5],
        sensitive_attributes={"foo": sf_data},
    )

    assert len(r) == 1
    assert numpy.isclose(r, 0.666666666)


def test_eod():
    r = fairical.metrics.calculate_metric(
        "eod+foo",
        y_true,
        y_pred,
        thresholds=[0.5],
        sensitive_attributes={"foo": sf_data},
    )

    assert len(r) == 1
    assert numpy.isclose(r, 1.00)


def test_eor():
    r = fairical.metrics.calculate_metric(
        "eor+foo",
        y_true,
        y_pred,
        thresholds=[0.5],
        sensitive_attributes={"foo": sf_data},
    )

    assert len(r) == 1
    assert numpy.isclose(r, 0.0)


def test_minmaxd_recall():
    r = fairical.metrics.calculate_metric(
        "minmaxd+rec+foo",
        y_true,
        y_pred,
        thresholds=[0.5],
        sensitive_attributes={"foo": sf_data},
    )

    assert len(r) == 1
    assert numpy.isclose(r, 0.1999999)


def test_minmaxr_recall():
    r = fairical.metrics.calculate_metric(
        "minmaxr+rec+foo",
        y_true,
        y_pred,
        thresholds=[0.5],
        sensitive_attributes={"foo": sf_data},
    )

    assert len(r) == 1
    assert numpy.isclose(r, 0.666667)


def test_should_minimize():
    for k in typing.get_args(fairical.metrics.UtilityMetricsType):
        if k in ("fnr", "fpr"):
            assert fairical.metrics.should_minimize(k)
        else:
            assert not fairical.metrics.should_minimize(k)

    for k in typing.get_args(fairical.metrics.FairnessMetricsType):
        if k in ("eod", "dpd"):
            assert fairical.metrics.should_minimize(k + "+foo")
        else:
            assert not fairical.metrics.should_minimize(k + "+foo")

    for k in typing.get_args(fairical.metrics.MinMaxFairnessMetricsType):
        if k == "minmaxd":
            assert fairical.metrics.should_minimize(k + "+rec+foo")
        else:
            assert not fairical.metrics.should_minimize(k + "+roc_auc+foo")



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import matplotlib

# avoids loading X11-dependent backends
matplotlib.use("agg")

import matplotlib.axes
import matplotlib.figure
import numpy
import pytest

from fairical import plot
from fairical.solutions import Solutions
from fairical.utils import IndicatorType


def test_radar_chart_output_type() -> None:
    indicators: dict[str, dict[IndicatorType, float]] = {
        "System 1": {
            "hv": 0.8,
            "ud": 0.3,
            "as": 0.4,
            "onvg": 10,
            "onvgr": 0.7,
            "relative-onvg": 4,
        },
        "System 2": {
            "hv": 0.6,
            "ud": 0.4,
            "as": 0.2,
            "onvg": 12,
            "onvgr": 0.8,
            "relative-onvg": 5,
        },
    }

    # radar_chart expects axes_keys to be subset of keys in dicts
    fig, ax = plot.radar_chart(indicators)
    assert isinstance(fig, matplotlib.figure.Figure)
    assert isinstance(ax, matplotlib.axes.Axes)


def test_radar_chart_exceptions() -> None:
    indicators: dict[str, dict[IndicatorType, float]] = {
        "System 1": {
            "hv": 0.8,
            "ud": 0.3,
            "as": 0.4,
            "onvg": 10,
            "onvgr": 0.7,
            "relative-onvg": 4,
        },
    }

    # axes too short
    with pytest.raises(AssertionError):
        plot.radar_chart(indicators, axes={"hv": ""})

    # invalid axes keys
    with pytest.raises(ValueError):
        plot.radar_chart(indicators, axes={"hv": "", "os": "", "foobar": ""})


def test_pareto_plot_errors() -> None:
    rng = numpy.random.default_rng()
    nds = Solutions.fromarray(rng.random((3, 1)), ("eod+race",), [0.1, 0.2, 0.3])
    ds = Solutions.fromarray(rng.random((2, 1)), ("eod+race",), [0.1, 0.2, 0.3])

    solutions = {"System 1": (nds, ds)}

    with pytest.raises(AssertionError):
        plot.pareto_plot(solutions, axes_labels={"acc": ""})


def test_pareto_2d_plot() -> None:
    rng = numpy.random.default_rng()

    solutions = {
        "System 1": (
            Solutions.fromarray(
                rng.random((3, 2)), ("eod+race", "acc"), [0.1, 0.2, 0.3]
            ),
            Solutions.fromarray(rng.random((2, 2)), ("eod+race", "acc"), [0.1, 0.2]),
        ),
        "System 2": (
            Solutions.fromarray(
                rng.random((3, 2)), ("eod+race", "acc"), [0.1, 0.2, 0.3]
            ),
            Solutions.fromarray(rng.random((2, 2)), ("eod+race", "acc"), [0.1, 0.2]),
        ),
    }

    fig, ax = plot.pareto_plot(
        solutions, axes_labels={"eod+race": "Race", "acc": "Acc"}
    )

    assert isinstance(fig, matplotlib.figure.Figure)
    assert isinstance(ax, matplotlib.axes.Axes)


def test_pareto_3d_plot() -> None:
    rng = numpy.random.default_rng()

    solutions = {
        "System 1": (
            Solutions.fromarray(
                rng.random((3, 3)), ("eod+race", "eod+age", "acc"), [0.1, 0.2, 0.3]
            ),
            Solutions.fromarray(
                rng.random((2, 3)), ("eod+race", "eod+age", "acc"), [0.1, 0.2]
            ),
        ),
        "System 2": (
            Solutions.fromarray(
                rng.random((3, 3)), ("eod+race", "eod+age", "acc"), [0.1, 0.2, 0.3]
            ),
            Solutions.fromarray(
                rng.random((2, 3)), ("eod+race", "eod+age", "acc"), [0.1, 0.2]
            ),
        ),
    }

    fig, ax = plot.pareto_plot(
        solutions, axes_labels={"eod+race": "Race", "eod+age": "Age", "acc": "Acc"}
    )

    assert isinstance(fig, matplotlib.figure.Figure)
    assert isinstance(ax, matplotlib.axes.Axes)



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Plotting utilities."""

import itertools
import typing

import matplotlib.axes
import matplotlib.colors
import matplotlib.figure
import matplotlib.pyplot
import numpy

from .metrics import FairnessMetricsType, MinMaxFairnessMetricsType, UtilityMetricsType
from .solutions import Solutions
from .utils import IndicatorType, extend_indicators, parse_indicator


def radar_chart(
    indicators: dict[str, dict[IndicatorType, float]],
    axes: dict[IndicatorType | str, str] = {
        "relative-onvg": r"$\widehat{ONVG}$",
        "onvgr": r"$ONVGR$",
        "ud": r"$UD$",
        "as": r"$AS$",
        "hv": r"$HV$",
    },
) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Generate radar chart for all systems under comparison.

    This method generates radar chart given performance indicator values in comparison
    of systems. It requires the presence of ``complement-ud`` and ``relative-onvg`` on
    ``indicators``.

    Parameters
    ----------
    indicators
        Indicators organized in a single dictionary where keys represent system labels,
        and values, dictionaries with *at least* the same keys as listed in
        ``axes_keys``.
    axes
        A dictionary containing the indicator keys that will be drawn on the radar
        chart, and corresponding labels associated with each of those axes. You can use
        LaTeX symbols and notations on the values of the dictionary.
    title
        The plot title.
    **kwargs
        Additional keyword arguments for updating chart properties. Supported options:
        - linewidth: Line width
        - linestyle: Line style

    Returns
    -------
        A tuple containing both the matplotlib figure and axes used to create the radar
        chart.
    """

    _hatch_list = ["/", "\\", "|", "-", "+", "x", "o", "O", ".", "*"]
    _hatch_cycle = itertools.cycle(_hatch_list)

    ndim = len(axes)

    # validate inputs
    assert ndim >= 3
    axes_ind: dict[IndicatorType, str] = {
        parse_indicator(k): v for k, v in axes.items()
    }

    extend_indicators(list(indicators.values()), list(axes_ind.keys()))

    values = numpy.array(
        [[v[k] for k in axes_ind.keys()] for v in indicators.values()], dtype=float
    )

    # concatenate the first column to last for looping
    values = numpy.column_stack((values, values[:, 0]))

    # draw plot
    angles = numpy.linspace(0, 2 * numpy.pi, values.shape[1], endpoint=True)
    fig, ax = matplotlib.pyplot.subplots(subplot_kw=dict(polar=True))
    labels = list(indicators.keys())
    areas = [k["area"] for k in indicators.values()]

    for i, value in enumerate(values):
        # draws the line around each system entry
        (line,) = ax.plot(
            angles, value, label=f"{labels[i]} ($\\triangle={areas[i]:.2f})$"
        )

        # fills the polygon
        ax.fill(angles, value, color=line.get_color(), alpha=0.25)

        # draws the hatching over the radar surface
        ax.fill(
            angles,
            value,
            color="none",
            edgecolor=line.get_color(),
            alpha=0.40,
            hatch=next(_hatch_cycle),
        )

    ax.set_xticks(angles[:-1])
    ax.set_xticklabels(axes.values())
    ax.set_yticks([0.2, 0.4, 0.6, 0.8, 1.0])
    ax.set_yticklabels(["0.2", "0.4", "0.6", "0.8", "1.0"])
    ax.tick_params(axis="both", which="major")
    ax.set_ylim(0, 1)

    ax.legend(
        loc="lower center",
        bbox_to_anchor=(0.5, -0.2),  # “where that corner goes in axes coords”
        bbox_transform=ax.transAxes,
        # ncol=len(indicators),  # spread entries horizontally
    )

    fig.tight_layout()

    return fig, ax


_AXES_LABELS_UTILITY: dict[UtilityMetricsType, str] = {
    "fpr": "Utility (FPR)",
    "tpr": "Utility (TPR)",
    "fnr": "Utility (FNR)",
    "tnr": "Utility (TNR)",
    "roc_auc": "Utility (AUROC)",
    "prec": "Utility (Precision)",
    "rec": "Utility (Recall)",
    "avg_prec": "Utility (Avg.Precision)",
    "f1": "Utility (F1-score)",
    "acc": "Utility (Accuracy)",
    "bal_acc": "Utility (Accuracy)",
}
"""Labels for pareto plots (utility)."""

_AXES_LABELS_FAIRNESS: dict[FairnessMetricsType, str] = {
    "dpd": "Fairness ($\\text{DPD}_\\text{%s})$",
    "dpr": "Fairness ($\\text{DPR}_\\text{%s})$",
    "eod": "Fairness ($\\text{EOD}_\\text{%s})$",
    "eor": "Fairness ($\\text{EOR}_\\text{%s})$",
}
"""Labels for pareto plots (fairness)."""

_AXES_LABELS_MINMAX_FAIRNESS: dict[MinMaxFairnessMetricsType, str] = {
    "minmaxd": "Min-Max Fairness (Diff. $\\text{%s}_\\text{%s})$",
    "minmaxr": "Min-Max Fairness (Ratio $\\text{%s}_\\text{%s})$",
}
"""Labels for pareto plots (min-max fairness)."""


def pareto_plot(
    solutions: dict[str, tuple[Solutions, Solutions]],
    axes_labels: dict[str, str] = {},
    alpha: float = 0.2,
    hide_ds: bool = False,
) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Generate pareto plot for all systems under comparison.

    This method generates pareto plot given solutions of
    systems in comparison.

    Parameters
    ----------
    solutions
        A dictionary where keys represent system names (that will be used as labels),
        and values are tuples with non-dominated (nds) and dominated solutions (ds)
        respectively.
    axes_labels
        If specified, overwrites the default labels for dimensions in
        :py:class:`fairical.solutions.Solutions`. Should be a dictionary that maps the
        keys in each :py:class:`fairical.solutions.Solutions` object to a single label.
        If not set, then we use a default setup provided in the module.
    alpha
        Alpha blend between non-dominated (fully opaque) and dominated solutions (partly
        transparent).
    hide_ds
        If true, hide the ds points for a-priori data from the plot.

    Returns
    -------
        A tuple of lists containing both the matplotlib figures and axes used to create the pareto
        plot. The lists will contain 2 elements each if `separate` is True, 1 otherwise.
    """

    _marker_list = ["o", "s", "^", "v", "<", ">", "d", "P", "X", "*", "+"]
    _marker_cycle = itertools.cycle(_marker_list)

    ndim = next(iter(solutions.values()))[0].n_metrics()

    # validate inputs
    assert 2 <= ndim <= 3

    def plot_pareto_front(nds_arr, ax, line_color):
        if nds_arr.shape[1] == 2:
            # sort by the first metric (x-axis) to get a sensible path
            order = numpy.argsort(nds_arr[:, 0])
            x_sorted, y_sorted = nds_arr[order].T
            ax.plot(
                x_sorted,
                y_sorted,
                linestyle="-",
                color=matplotlib.colors.to_rgba(line_color, alpha=min(2 * alpha, 1.0)),
                marker=marker,
                markerfacecolor=matplotlib.colors.to_rgba(line_color, alpha=1.0),
                linewidth=1,
                label=label,
            )
        else:
            # 3 dimensional case
            ax.plot_trisurf(
                *nds_arr.T,
                antialiased=False,
                shade=0,
                alpha=0.5,
                edgecolor="none",
            )

        return ax

    def resolve_ax_labels(solutions, axes_labels):
        # resolve which axes labels to use

        use_axes_labels: list[str] = []

        for k in next(iter(solutions.values()))[0].keys():
            if k in axes_labels:
                use_axes_labels.append(axes_labels[k])
            else:
                parts = k.split("+", 2)

                if parts[0] in typing.get_args(UtilityMetricsType):
                    use_axes_labels.append(
                        _AXES_LABELS_UTILITY[typing.cast(UtilityMetricsType, parts[0])]
                    )

                elif parts[0] in typing.get_args(FairnessMetricsType):
                    use_axes_labels.append(
                        _AXES_LABELS_FAIRNESS[
                            typing.cast(FairnessMetricsType, parts[0])
                        ]
                    )
                    if "%s" in use_axes_labels[-1]:
                        use_axes_labels[-1] = use_axes_labels[-1] % parts[1]

                elif parts[0] in typing.get_args(MinMaxFairnessMetricsType):
                    use_axes_labels.append(
                        _AXES_LABELS_MINMAX_FAIRNESS[
                            typing.cast(MinMaxFairnessMetricsType, parts[0])
                        ]
                    )
                    if "%s" in use_axes_labels[-1]:
                        use_axes_labels[-1] = use_axes_labels[-1] % tuple(parts[1:])
        return use_axes_labels

    fig = matplotlib.pyplot.figure()
    ax = fig.add_subplot(projection="3d") if ndim == 3 else fig.add_subplot()

    for label, (nds, ds) in solutions.items():
        marker = next(_marker_cycle)

        # plot non-dominated points with no transparency
        nds_arr = numpy.asarray(nds)
        pc_nds = ax.scatter(
            *nds_arr.T,
            marker=marker,
        )

        fc = pc_nds.get_facecolors()
        the_color = fc[0] if len(fc) else pc_nds.get_edgecolors()[0]

        if not hide_ds:
            # plot dominated points using the same marker and color
            ds_arr = numpy.asarray(ds)
            ax.scatter(
                *ds_arr.T,
                color=the_color,
                alpha=alpha,
                marker=marker,
            )

        plot_pareto_front(nds_arr, ax, the_color)

    use_axes_labels = resolve_ax_labels(solutions, axes_labels)

    ax.set_xlabel(use_axes_labels[0])
    ax.set_ylabel(use_axes_labels[1])
    if ndim == 3:
        ax.set_zlabel(use_axes_labels[2])

    ax.grid()
    ax.legend(
        loc="lower center",
        bbox_to_anchor=(0.5, -0.3),  # “where that corner goes in axes coords”
        bbox_transform=ax.transAxes,
    )

    fig.tight_layout()

    return fig, ax



# SPDX-FileCopyrightText: Copyright © 2025 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Shared utilities."""

import typing

import numpy
import tabulate

IndicatorType: typing.TypeAlias = typing.Literal[
    "hv",  # Hypervolume
    "ud",  # Uniform Distribution
    "os",  # Overall Pareto Spread
    "as",  # Average Pareto Spread
    "onvg",  # Overall Nondominated Vector Generation
    "onvgr",  # Overall Nondominated Vector Generation Ratio
    "relative-onvg",  # Relative ONVG
    "area",  # Radar chart area
]
"""Supported indicators type for pareto front estimates."""


def parse_indicator(ind: str | IndicatorType) -> IndicatorType:
    """Parse indicator from string.

    Parameters
    ----------
    ind
        Indicator prototype to parse.

    Returns
    -------
        Parsed indicator value.
    """
    allowed = typing.get_args(IndicatorType)

    if ind not in allowed:
        raise ValueError(f"Unknown indicator type: {ind!r}.  Must be one of {allowed}")

    return typing.cast(IndicatorType, ind)


def normalize_onvg(values: list[int]) -> list[float]:
    """Normalize ONVG indicators so range is :math:`[0, 1]`.

    This function normalizes the ONVG indicators of multiple systems so that it
    represents a ratio between the original value and the maximum for all systems
    compared.

    Parameters
    ----------
    values
        The values of all ONVG indicators to normalize.

    Returns
    -------
        A new list containing the values of indicators divided by their maximum.
    """

    return [k / max(values) for k in values]


def normalized_radar_area(values: list[int | float], maximum: float = 1.0) -> float:
    r"""Evaluate the radar-chart area formed by indicators of interest.

    This method calculates the "normalized" area (value between :math:`[0, 1]`) of a
    radar chart formed by indicators listed in ``values``.

    An intuitive way to calculate the area of radar chart is to consider it as a set of
    triangles, defined by the chart axes and the angles between then, out of which you
    know the sizes of two sides, which are given, and the angle between them, which is
    fixed (:math:`2\pi/n`). For example, the area of a 3-way radar chart is therefore
    the total area of 3 triangles with sides equal to each combination of input
    ``values``, with angles of :math:`120^o`.  More generally, the total area can be
    defined as:

    .. math::

       \sum_i^n 0.5 a_i b_i \sin(2\pi/n)

    Where :math:`n` is the total number of axes on the radar chart, and `a_i` and `b_i`
    are the adjacent axes for which we are computing the section area. To normalize this
    such that all charts have a maximum area of 1.0, one must bind the maximum values in
    each of the radar chart axes.  In this implementation, we bind these maxima to 1.0.
    With that, one can compute the largest radar chart area and normalize the given area
    by that value.

    If one considers each triangle individually, it becomes clear that the factor
    :math:`0.5 \sin(2\pi/n)` cancels out and only :math:`a b / max^2` matters.
    This simplified version is implemented here for maximum accuracy and speed.

    Parameters
    ----------
    values
        The values of the radar chart plot. Naturally, at least 3 values must be
        provided.  All values are required to lie in the interval :math:`[0, 1]`.
    maximum
        The maximum value one can have in each axis of the radar chart.  This value is
        used to compute the normalization factor.

    Returns
    -------
        The "normalized" area (value between :math:`[0, 1]`) of a
        radar chart formed by indicators listed in ``values``.
    """

    def _norm_triangle_area(a, b):
        return a * b / (maximum**2)

    assert len(values) >= 3, "At least 3 values are required."
    assert all([0 <= k <= maximum for k in values])

    return sum(
        [_norm_triangle_area(*k) for k in zip(values, values[1:] + values[:1])]
    ) / len(values)


def extend_indicators(
    indicators: typing.Sequence[dict[IndicatorType, float]],
    radar_axes: typing.Sequence[IndicatorType] = [
        "relative-onvg",
        "onvgr",
        "ud",
        "as",
        "hv",
    ],
) -> None:
    """Extend indicators of each system with relative metrics.

    This method adds ``relative-onvg``, and relative radar chart area on ``area``.  The
    radar chart area is calculated based on the axes selected on ``radar_axes``.

    Note this function **modifies the indicator dictionaries in-place**.

    Parameters
    ----------
    indicators
        Indicators organized in a dictionary of dictionaries where keys represent the
        labels of each system, and values, dictionaries that represent indicators for
        that system with *at least* keys listed in ``table_keys``. We assume the
        following metrics are calculated for every system:

        * ``hv``: the pareto estimate hypervolume (float)
        * ``onvg``: the number of non-dominated solutions (int)
        * ``onvgr``: the ratio between the number of non-dominated solutions and the
          total number of solutions (int)
        * ``ud``: the uniformity of non-dominated solutions across the estimated front
          (float)
        * ``as``: the average spread of non-dominated solutions across the estimated
          front (float)

    radar_axes
        The indicator keys that will be used for estimating the normalized radar surface
        for each system.
    """

    ordered_onvg = [typing.cast(int, k["onvg"]) for k in indicators]
    for v, rel_onvg in zip(indicators, normalize_onvg(ordered_onvg)):
        v["relative-onvg"] = rel_onvg
        v["area"] = normalized_radar_area(
            [typing.cast(float, v[parse_indicator(k)]) for k in radar_axes]
        )


def make_table(
    indicators: dict[str, dict[IndicatorType, float]],
    table_keys: typing.Sequence[IndicatorType | str] = [
        "relative-onvg",
        "onvgr",
        "ud",
        "as",
        "hv",
    ],
    fmt: str = "simple",
) -> str:
    """Extract and format table from pre-computed evaluation data.

    Extracts elements from ``data`` that can be displayed on a
    terminal-style table, format, and return it.

    Parameters
    ----------
    indicators
        Indicators organized in a dictionary of dictionaries where keys represent the
        labels of each system, and values, dictionaries that represent indicators for
        that system with *at least* keys listed in ``table_keys``. We assume the
        following metrics are calculated for every system:

        * ``hv``: the pareto estimate hypervolume (float)
        * ``onvg``: the number of non-dominated solutions (int)
        * ``onvgr``: the ratio between the number of non-dominated solutions and the
          total number of solutions (int)
        * ``ud``: the uniformity of non-dominated solutions across the estimated front
          (float)
        * ``as``: the average spread of non-dominated solutions across the estimated
          front (float)

    table_keys
        The indicator keys that will be tabulated in the table.
    fmt
        One of the formats supported by `python-tabulate
        <https://pypi.org/project/tabulate/>`_. Default is "github".

    Returns
    -------
        A string representation of a table.
    """

    table_keys_ind: list[IndicatorType] = [parse_indicator(k) for k in table_keys]

    extend_indicators(list(indicators.values()), table_keys_ind)

    table_headers = ["System"] + [k.upper() for k in table_keys] + ["Area"]

    values = numpy.array(
        [[v[k] for k in table_keys_ind + ["area"]] for v in indicators.values()],
        dtype=float,
    )

    table_data = []
    for system_name, system_data in zip(indicators.keys(), values):
        table_data.append([system_name] + system_data.tolist())

    return tabulate.tabulate(
        table_data, table_headers, tablefmt=fmt, floatfmt=".2f", stralign="right"
    )
